# Copyright 2023 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Address, IPv4Network, IPv6Network
from typing import List, Literal, Optional, Set
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, IPvAnyAddress, field_validator, model_validator

from catalystwan.models.common import InterfaceType, TLOCColor, check_fields_exclusive


def check_jitter_ms(jitter_str: str) -> str:
    assert 1 <= int(jitter_str) <= 1000
    return jitter_str


def check_latency_ms(latency_str: str) -> str:
    assert 1 <= int(latency_str) <= 1000
    return latency_str


def check_loss_percent(loss_str: str) -> str:
    assert 0 <= int(loss_str) <= 100
    return loss_str


PolicerExceedAction = Literal[
    "drop",
    "remark",
]

EncapType = Literal[
    "ipsec",
    "gre",
]

PathPreference = Literal[
    "direct-path",
    "multi-hop-path",
    "all-paths",
]


class ColorDSCPMap(BaseModel):
    color: TLOCColor
    dscp: int = Field(ge=0, le=63)


class ColorGroupPreference(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    color_preference: str = Field(serialization_alias="colorPreference", validation_alias="colorPreference")
    path_preference: PathPreference = Field(serialization_alias="pathPreference", validation_alias="pathPreference")

    @staticmethod
    def from_color_set_and_path(
        color_preference: Set[TLOCColor], path_preference: PathPreference
    ) -> "ColorGroupPreference":
        return ColorGroupPreference(color_preference=" ".join(color_preference), path_preference=path_preference)


class FallbackBestTunnel(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    criteria: str
    jitter_variance: Optional[str] = Field(
        default=None,
        serialization_alias="jitterVariance",
        validation_alias="jitterVariance",
        description="jitter variance in ms",
    )
    latency_variance: Optional[str] = Field(
        default=None,
        serialization_alias="latencyVariance",
        validation_alias="latencyVariance",
        description="latency variance in ms",
    )
    loss_variance: Optional[str] = Field(
        default=None,
        serialization_alias="lossVariance",
        validation_alias="lossVariance",
        description="loss variance as percentage",
    )
    _criteria_priority: List[Literal["jitter", "latency", "loss"]] = []

    # validators
    _jitter_validator = field_validator("jitter_variance")(check_jitter_ms)
    _latency_validator = field_validator("latency_variance")(check_latency_ms)
    _loss_validator = field_validator("loss_variance")(check_loss_percent)

    @model_validator(mode="after")
    def check_criteria(self):
        expected_criteria = set()
        if self.jitter_variance is not None:
            expected_criteria.add("jitter")
        if self.latency_variance is not None:
            expected_criteria.add("latency")
        if self.loss_variance is not None:
            expected_criteria.add("loss")
        assert expected_criteria, "At least one variance type needs to be present"
        self._criteria_priority = str(self.criteria).split("-")
        observed_criteria = set(self._criteria_priority)
        assert expected_criteria == observed_criteria
        return self

    def _update_criteria_field(self) -> None:
        self.criteria = f"{'-'.join(self._criteria_priority)}"

    def add_jitter_criteria(self, jitter_variance: int) -> None:
        if self.jitter_variance is None:
            self._criteria_priority.append("jitter")
        self.jitter_variance = str(jitter_variance)
        self._update_criteria_field()
        self.check_criteria

    def add_latency_criteria(self, latency_variance: int) -> None:
        if self.latency_variance is None:
            self._criteria_priority.append("latency")
        self.latency_variance = str(latency_variance)
        self._update_criteria_field()
        self.check_criteria

    def add_loss_criteria(self, loss_variance: int) -> None:
        if self.loss_variance is None:
            self._criteria_priority.append("loss")
        self.loss_variance = str(loss_variance)
        self._update_criteria_field()
        self.check_criteria


class DataPrefixListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    ip_prefix: IPv4Network = Field(serialization_alias="ipPrefix", validation_alias="ipPrefix")


class SiteListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    site_id: str = Field(serialization_alias="siteId", validation_alias="siteId")


class VPNListEntry(BaseModel):
    vpn: str = Field(description="0-65530 range or single number")

    @field_validator("vpn")
    @classmethod
    def check_vpn_range(cls, vpns_str: str):
        vpns = [int(vpn) for vpn in vpns_str.split("-")]
        assert len(vpns) <= 2
        for vpn in vpns:
            assert 0 <= vpn <= 65530
        if len(vpns) == 2:
            assert vpns[0] <= vpns[1]
        return vpns_str


class ZoneListEntry(BaseModel):
    vpn: Optional[str] = Field(default=None, description="0-65530 single number")
    interface: Optional[InterfaceType] = None

    @field_validator("vpn")
    @classmethod
    def check_vpn_range(cls, vpn_str: str):
        assert 0 <= int(vpn_str) <= 65530
        return vpn_str

    @model_validator(mode="after")
    def check_vpn_xor_interface(self):
        check_fields_exclusive(self.__dict__, {"vpn", "interface"}, True)
        return self


class FQDNListEntry(BaseModel):
    pattern: str


class GeoLocationListEntry(BaseModel):
    country: Optional[str] = Field(default=None, description="ISO-3166 alpha-3 country code eg: FRA")
    continent: Optional[str] = Field(
        default=None, description="One of 2-letter continent codes: AF, NA, OC, AN, AS, EU, SA"
    )

    @model_validator(mode="after")
    def check_country_xor_continent(self):
        check_fields_exclusive(self.__dict__, {"country", "continent"}, True)
        return self


class PortListEntry(BaseModel):
    port: str

    @field_validator("port")
    @classmethod
    def check_port_range(cls, port_str: str):
        assert 0 <= int(port_str) <= 65535
        return port_str


class ProtocolNameListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    protocol_name: str = Field(serialization_alias="protocolName", validation_alias="protocolName")


class LocalAppListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    app_family: Optional[str] = Field(default=None, serialization_alias="appFamily", validation_alias="appFamily")
    app: Optional[str] = None

    @model_validator(mode="after")
    def check_app_xor_appfamily(self):
        check_fields_exclusive(self.__dict__, {"app", "app_family"}, True)
        return self


class AppListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    app_family: Optional[str] = Field(default=None, serialization_alias="appFamily", validation_alias="appFamily")
    app: Optional[str] = None

    @model_validator(mode="after")
    def check_app_xor_appfamily(self):
        check_fields_exclusive(self.__dict__, {"app", "app_family"}, True)
        return self


class ColorListEntry(BaseModel):
    color: TLOCColor


class DataIPv6PrefixListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    ipv6_prefix: IPv6Network = Field(serialization_alias="ipv6Prefix", validation_alias="ipv6Prefix")


class LocalDomainListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    name_server: str = Field(
        pattern="^[^*+].*",
        serialization_alias="nameServer",
        validation_alias="nameServer",
        max_length=240,
        description="Must be valid std regex."
        "String cannot start with a '*' or a '+', be empty, or be more than 240 characters",
    )


class IPSSignatureListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    generator_id: str = Field(serialization_alias="generatorId", validation_alias="generatorId")
    signature_id: str = Field(serialization_alias="signatureId", validation_alias="signatureId")


class URLListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    pattern: str


class CommunityListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    community: str = Field(description="Example: 1000:10000 or internet or local-AS or no advertise or no-export")


class PolicerListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    burst: str = Field(description="bytes: integer in range 15000-10000000")
    exceed: PolicerExceedAction = "drop"
    rate: str = Field(description="bps: integer in range 8-100000000000")

    @field_validator("burst")
    @classmethod
    def check_burst(cls, burst_str: str):
        assert 15000 <= int(burst_str) <= 10_000_000
        return burst_str

    @field_validator("rate")
    @classmethod
    def check_rate(cls, rate_str: str):
        assert 8 <= int(rate_str) <= 100_000_000_000
        return rate_str


class ASPathListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    as_path: str = Field(serialization_alias="asPath", validation_alias="asPath")


class ClassMapListEntry(BaseModel):
    queue: str

    @field_validator("queue")
    @classmethod
    def check_queue(cls, queue_str: str):
        assert 0 <= int(queue_str) <= 7
        return queue_str


class MirrorListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    remote_dest: IPvAnyAddress = Field(serialization_alias="remoteDest", validation_alias="remoteDest")
    source: IPvAnyAddress


class AppProbeClassListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    forwarding_class: str = Field(serialization_alias="forwardingClass", validation_alias="forwardingClass")
    map: List[ColorDSCPMap] = []

    def add_color_mapping(self, color: TLOCColor, dscp: int) -> None:
        self.map.append(ColorDSCPMap(color=color, dscp=dscp))


class SLAClassListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    latency: Optional[str] = None
    loss: Optional[str] = None
    jitter: Optional[str] = None
    app_probe_class: Optional[UUID] = Field(serialization_alias="appProbeClass", validation_alias="appProbeClass")
    fallback_best_tunnel: Optional[FallbackBestTunnel] = Field(
        default=None, serialization_alias="fallbackBestTunnel", validation_alias="fallbackBestTunnel"
    )

    # validators
    _jitter_validator = field_validator("jitter")(check_jitter_ms)
    _latency_validator = field_validator("latency")(check_latency_ms)
    _loss_validator = field_validator("loss")(check_loss_percent)

    @model_validator(mode="after")
    def check_at_least_one_criteria_is_set(self):
        assert any([self.latency, self.loss, self.jitter])
        return self

    def add_fallback_jitter_criteria(self, jitter_variance: int) -> None:
        if self.fallback_best_tunnel:
            self.fallback_best_tunnel.add_jitter_criteria(jitter_variance)
        else:
            self.fallback_best_tunnel = FallbackBestTunnel(criteria="jitter", jitter_variance=str(jitter_variance))

    def add_fallback_latency_criteria(self, latency_variance: int) -> None:
        if self.fallback_best_tunnel:
            self.fallback_best_tunnel.add_latency_criteria(latency_variance)
        else:
            self.fallback_best_tunnel = FallbackBestTunnel(criteria="latency", latency_variance=str(latency_variance))

    def add_fallback_loss_criteria(self, loss_variance: int) -> None:
        if self.fallback_best_tunnel:
            self.fallback_best_tunnel.add_loss_criteria(loss_variance)
        else:
            self.fallback_best_tunnel = FallbackBestTunnel(criteria="loss", loss_variance=str(loss_variance))


class TLOCListEntry(BaseModel):
    tloc: IPv4Address
    color: TLOCColor
    encap: EncapType
    preference: Optional[str] = None

    @field_validator("preference")
    @classmethod
    def check_preference(cls, preference_str: str):
        if preference_str is not None:
            assert 0 <= int(preference_str) <= 2**32 - 1
            return preference_str


class PreferredColorGroupListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    primary_preference: ColorGroupPreference = Field(
        serialization_alias="primaryPreference", validation_alias="primaryPreference"
    )
    secondary_preference: Optional[ColorGroupPreference] = Field(
        default=None, serialization_alias="secondaryPreference", validation_alias="secondaryPreference"
    )
    tertiary_preference: Optional[ColorGroupPreference] = Field(
        default=None, serialization_alias="tertiaryPreference", validation_alias="tertiaryPreference"
    )

    @model_validator(mode="after")
    def check_optional_preferences_order(self):
        assert not (self.secondary_preference is None and self.tertiary_preference is not None)
        return self


class PrefixListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    ip_prefix: IPv4Network = Field(serialization_alias="ipPrefix", validation_alias="ipPrefix")
    ge: Optional[str] = None
    le: Optional[str] = None

    @field_validator("ge", "le", check_fields=False)
    @classmethod
    def check_ge_and_le(cls, ge_le_str: Optional[str]):
        if ge_le_str is not None:
            assert 0 <= int(ge_le_str) <= 32
        return ge_le_str


class IPv6PrefixListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    ipv6_prefix: IPv6Network = Field(serialization_alias="ipv6Prefix", validation_alias="ipv6Prefix")
    ge: Optional[str] = None
    le: Optional[str] = None

    @field_validator("ge", "le", check_fields=False)
    @classmethod
    def check_ge_and_le(cls, ge_le_str: Optional[str]):
        if ge_le_str is not None:
            assert 0 <= int(ge_le_str) <= 128
        return ge_le_str


class RegionListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    region_id: str = Field(
        serialization_alias="regionId", validation_alias="regionId", description="Number in range 0-63"
    )

    @field_validator("region_id")
    @classmethod
    def check_region_id(cls, region_id_str: str):
        regions = [int(region_id) for region_id in region_id_str.split("-")]
        assert len(regions) <= 2
        for region in regions:
            assert 0 <= region <= 63
        if len(regions) == 2:
            assert regions[0] <= regions[1]
        return region_id_str



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable
from catalystwan.models.configuration.feature_profile.common import Prefix

ProtocolIPv4 = Literal[
    "bgp",
    "ospf",
    "opsfv3",
    "connected",
    "static",
    "network",
    "aggregate",
    "eigrp",
    "lisp",
    "isis",
]

ProtocolIPv6 = Literal[
    "BGP",
    "OSPF",
    "connected",
    "static",
    "network",
    "aggregate",
]

Region = Literal[
    "core-and-access",
    "core",
    "access",
]

NATRoute = Literal[
    "NAT64",
    "NAT66",
]

ServiceType = Literal[
    "FW",
    "IDS",
    "IDP",
    "netsvc1",
    "netsvc2",
    "netsvc3",
    "netsvc4",
    "TE",
    "appqoe",
]

ServiceRouteType = Literal[
    "SIG",
    "SSE",
]

Direction = Literal[
    "inside",
    "outside",
]

NATPortForwardProtocol = Literal[
    "TCP",
    "UDP",
]

RouteLeakFromGlobalProtocol = Literal[
    "static",
    "connected",
    "bgp",
    "ospf",
]

RedistributeToServiceProtocol = Literal[
    "bgp",
    "ospf",
]

RouteLeakFromServiceProtocol = Literal[
    "static",
    "connected",
    "bgp",
    "ospf",
]

RedistributeToGlobalProtocol = Literal[
    "bgp",
    "ospf",
]


class DnsIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    primary_dns_address_ipv4: Union[Variable, Global[str], Default[None]] = Field(
        serialization_alias="primaryDnsAddressIpv4", validation_alias="primaryDnsAddressIpv4"
    )
    secondary_dns_address_ipv4: Union[Variable, Global[str], Default[None]] = Field(
        serialization_alias="secondaryDnsAddressIpv4", validation_alias="secondaryDnsAddressIpv4"
    )


class DnsIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    primary_dns_address_ipv6: Union[Variable, Global[str], Default[None]] = Field(
        serialization_alias="primaryDnsAddressIpv6", validation_alias="primaryDnsAddressIpv6"
    )
    secondary_dns_address_ipv6: Union[Variable, Global[str], Default[None]] = Field(
        serialization_alias="secondaryDnsAddressIpv6", validation_alias="secondaryDnsAddressIpv6"
    )


class HostMapping(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    host_name: Union[Variable, Global[str]] = Field(serialization_alias="hostName", validation_alias="hostName")
    list_of_ip: Union[Variable, Global[str]] = Field(serialization_alias="listOfIp", validation_alias="listOfIp")


class RoutePrefix(BaseModel):
    ip_address: Union[Variable, Global[str]] = Field(serialization_alias="ipAddress", validation_alias="ipAddress")
    subnet_mask: Union[Variable, Global[str]] = Field(serialization_alias="subnetMask", validation_alias="subnetMask")


class IPv4Prefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Prefix
    aggregate_only: Optional[Union[Global[bool], Default[bool]]] = Field(
        serialization_alias="aggregateOnly", validation_alias="aggregateOnly", default=None
    )
    region: Optional[Union[Variable, Global[Region], Default[str]]] = None


class IPv6Prefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Union[Global[str], Variable]
    aggregate_only: Optional[Union[Global[bool], Default[bool]]] = Field(
        serialization_alias="aggregateOnly", validation_alias="aggregateOnly", default=None
    )
    region: Optional[Union[Variable, Global[Region], Default[str]]] = None


class OmpAdvertiseIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    omp_protocol: Union[Variable, Global[ProtocolIPv4]] = Field(
        serialization_alias="ompProtocol", validation_alias="ompProtocol"
    )
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )
    prefix_list: Optional[List[IPv4Prefix]] = None


class OmpAdvertiseIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    omp_protocol: Union[Variable, Global[ProtocolIPv6]] = Field(
        serialization_alias="ompProtocol", validation_alias="ompProtocol"
    )
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )
    prefix_list: Optional[List[IPv6Prefix]] = None


class IPv4RouteGatewayNextHop(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Variable, Global[str]]
    distance: Union[Variable, Global[int], Default[int]] = Default[int](value=1)


class IPv4RouteGatewayNextHopWithTracker(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Variable, Global[str]]
    distance: Union[Variable, Global[int], Default[int]] = Default[int](value=1)
    tracker: Union[Global[UUID], Default[None]]


class NextHopContainer(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    next_hop: Optional[List[IPv4RouteGatewayNextHop]] = Field(
        serialization_alias="nextHop", validation_alias="nextHop", default=None
    )
    next_hop_with_tracker: Optional[List[IPv4RouteGatewayNextHopWithTracker]] = Field(
        serialization_alias="nextHopWithTracker", validation_alias="nextHopWithTracker", default=None
    )


class IPv6RouteGatewayNextHop(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Variable, Global[str]]
    distance: Union[Variable, Global[int], Default[int]] = Default[int](value=1)


class NextHopIPv6Container(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    next_hop: Optional[List[IPv6RouteGatewayNextHop]] = Field(
        serialization_alias="nextHop", validation_alias="nextHop", default=None
    )


class NextHopRouteContainer(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    next_hop_container: NextHopContainer = Field(
        serialization_alias="nextHopContainer", validation_alias="nextHopContainer"
    )


class NextHopRouteIPv6Container(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    next_hop_container: NextHopIPv6Container = Field(
        serialization_alias="nextHopContainer", validation_alias="nextHopContainer"
    )


class Null0(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    null0: Union[Global[bool], Default[bool]] = Default[bool](value=True)
    distance: Union[Variable, Global[int], Default[int]] = Default[int](value=1)


class Null0IPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    null0: Union[Global[bool], Default[bool]] = Default[bool](value=True)


class NAT(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    nat: Union[Variable, Global[NATRoute]]


class DHCP(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    dhcp: Union[Global[bool], Default[bool]] = Default[bool](value=True)


class StaticRouteVPN(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    vpn: Union[Global[bool], Default[bool]] = Default[bool](value=True)


class NextHopInterfaceRoute(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Variable, Global[str], Default[None]] = Default[None](value=None)
    distance: Union[Variable, Global[int], Default[int]] = Default[int](value=1)


class NextHopInterfaceRouteIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Variable, Global[str], Default[None]] = Default[None](value=None)
    distance: Union[Variable, Global[int], Default[int]] = Default[int](value=1)


class IPStaticRouteInterface(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Union[Variable, Global[str]] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    next_hop: List[NextHopInterfaceRoute] = Field(serialization_alias="nextHop", validation_alias="nextHop")


class IPv6StaticRouteInterface(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Union[Variable, Global[str]] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    next_hop: List[NextHopInterfaceRouteIPv6] = Field(serialization_alias="nextHop", validation_alias="nextHop")


class InterfaceContainer(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ip_static_route_interface: List[IPStaticRouteInterface] = Field(
        serialization_alias="ipStaticRouteInterface", validation_alias="ipStaticRouteInterface"
    )


class InterfaceIPv6Container(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ipv6_static_route_interface: List[IPv6StaticRouteInterface] = Field(
        serialization_alias="ipv6StaticRouteInterface", validation_alias="ipv6StaticRouteInterface"
    )


class InterfaceRouteContainer(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_container: InterfaceContainer = Field(
        serialization_alias="interfaceContainer", validation_alias="interfaceContainer"
    )


class InterfaceRouteIPv6Container(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_container: InterfaceIPv6Container = Field(
        serialization_alias="interfaceContainer", validation_alias="interfaceContainer"
    )


class StaticRouteIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: RoutePrefix
    one_of_ip_route: Union[NextHopRouteContainer, Null0, DHCP, StaticRouteVPN, InterfaceRouteContainer] = Field(
        serialization_alias="oneOfIpRoute", validation_alias="oneOfIpRoute"
    )


class StaticRouteIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: RoutePrefix
    one_of_ip_route: Union[NextHopRouteIPv6Container, Null0IPv6, NAT, InterfaceRouteIPv6Container] = Field(
        serialization_alias="oneOfIpRoute", validation_alias="oneOfIpRoute"
    )


class Service(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    service_type: Union[Variable, Global[ServiceType]] = Field(
        serialization_alias="serviceType", validation_alias="serviceType"
    )
    ipv4_addresses: Union[Variable, Global[List[str]]] = Field(
        serialization_alias="ipv4Addresses", validation_alias="ipv4Addresses"
    )
    tracking: Union[Variable, Global[bool], Default[bool]] = Default[bool](value=True)


class ServiceRoute(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Prefix
    service: Union[Variable, Global[ServiceRouteType], Default[ServiceRouteType]] = Default[ServiceRouteType](
        value="SIG"
    )
    vpn: Global[int] = Global[int](value=0)
    sse_instance: Optional[Union[Variable, Global[str]]] = Field(
        serialization_alias="sseInstance", validation_alias="sseInstance", default=None
    )


class StaticGreRouteIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Prefix
    interface: Union[Variable, Global[List[str]], Default[None]]
    vpn: Global[int] = Global[int](value=0)


class StaticIpsecRouteIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    prefix: Prefix
    interface: Union[Variable, Global[List[str]], Default[None]]


class NatPool(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    nat_pool_name: Union[Variable, Global[int]] = Field(
        serialization_alias="natPoolName", validation_alias="natPoolName"
    )
    prefix_length: Union[Variable, Global[int]] = Field(
        serialization_alias="prefixLength", validation_alias="prefixLength"
    )
    range_start: Union[Variable, Global[str]] = Field(serialization_alias="rangeStart", validation_alias="rangeStart")
    range_end: Union[Variable, Global[str]] = Field(serialization_alias="rangeEnd", validation_alias="rangeEnd")
    overload: Union[Variable, Global[bool], Default[bool]] = Default[bool](value=True)
    direction: Union[Variable, Global[Direction]]
    tracking_object: Optional[dict] = Field(
        serialization_alias="trackingObject", validation_alias="trackingObject", default=None
    )


class NatPortForward(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    nat_pool_name: Union[Variable, Global[int], Default[None]] = Field(
        serialization_alias="natPoolName", validation_alias="natPoolName"
    )
    source_port: Union[Variable, Global[int]] = Field(serialization_alias="sourcePort", validation_alias="sourcePort")
    translate_port: Union[Variable, Global[int]] = Field(
        serialization_alias="translatePort", validation_alias="translatePort"
    )
    source_ip: Union[Variable, Global[str]] = Field(serialization_alias="sourceIp", validation_alias="sourceIp")
    translated_source_ip: Union[Variable, Global[str]] = Field(
        serialization_alias="TranslatedSourceIp", validation_alias="TranslatedSourceIp"
    )
    protocol: Union[Variable, Global[NATPortForwardProtocol]]


class StaticNat(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    nat_pool_name: Union[Variable, Global[int], Default[None]] = Field(
        serialization_alias="natPoolName", validation_alias="natPoolName"
    )
    source_ip: Union[Variable, Global[str]] = Field(serialization_alias="sourceIp", validation_alias="sourceIp")
    translated_source_ip: Union[Variable, Global[str]] = Field(
        serialization_alias="TranslatedSourceIP", validation_alias="TranslatedSourceIP"
    )
    static_nat_direction: Union[Variable, Global[Direction]] = Field(
        serialization_alias="staticNatDirection", validation_alias="staticNatDirection"
    )
    tracking_object: Optional[dict] = Field(
        serialization_alias="trackingObject", validation_alias="trackingObject", default=None
    )


class StaticNatSubnet(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_ip_subnet: Union[Variable, Global[str]] = Field(
        serialization_alias="sourceIpSubnet", validation_alias="sourceIpSubnet"
    )
    translated_source_ip_subnet: Union[Variable, Global[str]] = Field(
        serialization_alias="TranslatedSourceIpSubnet", validation_alias="TranslatedSourceIpSubnet"
    )
    prefix_length: Union[Variable, Global[int]] = Field(
        serialization_alias="prefixLength", validation_alias="prefixLength"
    )
    static_nat_direction: Union[Variable, Global[Direction]] = Field(
        serialization_alias="staticNatDirection", validation_alias="staticNatDirection"
    )
    tracking_object: Optional[dict] = Field(
        serialization_alias="trackingObject", validation_alias="trackingObject", default=None
    )


class Nat64v4Pool(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    nat64_v4_pool_name: Union[Variable, Global[str]] = Field(
        serialization_alias="nat64V4PoolName", validation_alias="nat64V4PoolName"
    )
    nat64_v4_pool_range_start: Union[Variable, Global[str]] = Field(
        serialization_alias="nat64V4PoolRangeStart", validation_alias="nat64V4PoolRangeStart"
    )
    nat64_v4_pool_range_end: Union[Variable, Global[str]] = Field(
        serialization_alias="nat64V4PoolRangeEnd", validation_alias="nat64V4PoolRangeEnd"
    )
    nat64_v4_pool_overload: Union[Variable, Global[bool], Default[bool]] = Field(
        serialization_alias="nat64V4PoolOverload",
        validation_alias="nat64V4PoolOverload",
        default=Default[bool](value=False),
    )


class RedistributeToService(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    protocol: Union[Variable, Global[RedistributeToServiceProtocol]]
    policy: Union[Default[None], Global[UUID]]


class RedistributeToGlobal(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    protocol: Union[Variable, Global[RedistributeToGlobalProtocol]]
    policy: Union[Default[None], Global[UUID]]


class RouteLeakFromGlobal(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    route_protocol: Union[Variable, Global[RouteLeakFromGlobalProtocol]] = Field(
        serialization_alias="routeProtocol", validation_alias="routeProtocol"
    )
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )
    redistribute_to_protocol: Optional[List[RedistributeToService]] = Field(
        serialization_alias="redistributeToProtocol", validation_alias="redistributeToProtocol", default=None
    )


class RouteLeakFromService(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    route_protocol: Union[Variable, Global[RouteLeakFromServiceProtocol]] = Field(
        serialization_alias="routeProtocol", validation_alias="routeProtocol"
    )
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )
    redistribute_to_protocol: Optional[List[RedistributeToService]] = Field(
        serialization_alias="redistributeToProtocol", validation_alias="redistributeToProtocol", default=None
    )


class RouteLeakBetweenServices(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_vpn: Union[Variable, Global[int]] = Field(serialization_alias="soureVpn", validation_alias="soureVpn")
    route_protocol: Union[Variable, Global[RouteLeakFromServiceProtocol]] = Field(
        serialization_alias="routeProtocol", validation_alias="routeProtocol"
    )
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )
    redistribute_to_protocol: Optional[List[RedistributeToService]] = Field(
        serialization_alias="redistributeToProtocol", validation_alias="redistributeToProtocol", default=None
    )


class RouteTarget(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    rt: Union[Global[str], Variable]


class MplsVpnIPv4RouteTarget(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    import_rt_list: Optional[List[RouteTarget]] = Field(
        serialization_alias="importRtList", validation_alias="importRtList", default=None
    )
    export_rt_list: Optional[List[RouteTarget]] = Field(
        serialization_alias="exportRtList", validation_alias="exportRtList", default=None
    )


class MplsVpnIPv6RouteTarget(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    import_rt_list: Optional[List[RouteTarget]] = Field(
        serialization_alias="importRtList", validation_alias="importRtList", default=None
    )
    export_rt_list: Optional[List[RouteTarget]] = Field(
        serialization_alias="exportRtList", validation_alias="exportRtList", default=None
    )


class LanVpnData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    vpn_id: Union[Variable, Global[int], Default[int]] = Field(serialization_alias="vpnId", validation_alias="vpnId")
    name: Union[Variable, Global[str], Default[None]]
    omp_admin_distance: Optional[Union[Variable, Global[int], Default[None]]] = Field(
        serialization_alias="ompAdminDistance", validation_alias="ompAdminDistance", default=None
    )
    omp_admin_distance_ipv6: Optional[Union[Variable, Global[int], Default[None]]] = Field(
        serialization_alias="ompAdminDistanceIpv6", validation_alias="ompAdminDistanceIpv6", default=None
    )
    dns_ipv4: Optional[DnsIPv4] = Field(serialization_alias="dnsIpv4", validation_alias="dnsIpv4", default=None)
    dns_ipv6: Optional[DnsIPv6] = Field(serialization_alias="dnsIpv6", validation_alias="dnsIpv6", default=None)
    new_host_mapping: Optional[List[HostMapping]] = Field(
        serialization_alias="newHostMapping", validation_alias="newHostMapping", default=None
    )
    omp_advertise_ipv4: Optional[List[OmpAdvertiseIPv4]] = Field(
        serialization_alias="ompAdvertiseIpv4", validation_alias="ompAdvertiseIpv4", default=None
    )
    omp_advertise_ipv6: Optional[List[OmpAdvertiseIPv6]] = Field(
        serialization_alias="ompAdvertiseIpv6", validation_alias="ompAdvertiseIpv6", default=None
    )
    ipv4_route: Optional[List[StaticRouteIPv4]] = Field(
        serialization_alias="ipv4Route", validation_alias="ipv4Route", default=None
    )
    ipv6_route: Optional[List[StaticRouteIPv6]] = Field(
        serialization_alias="ipv6Route", validation_alias="ipv6Route", default=None
    )
    service: Optional[List[Service]] = None
    service_route: Optional[List[ServiceRoute]] = Field(
        serialization_alias="serviceRoute", validation_alias="serviceRoute", default=None
    )
    gre_route: Optional[List[StaticGreRouteIPv4]] = Field(
        serialization_alias="greRoute", validation_alias="greRoute", default=None
    )
    ipsec_route: Optional[List[StaticIpsecRouteIPv4]] = Field(
        serialization_alias="ipsecRoute", validation_alias="ipsecRoute", default=None
    )
    nat_pool: Optional[List[NatPool]] = Field(serialization_alias="natPool", validation_alias="natPool", default=None)
    nat_port_forwarding: Optional[List[NatPortForward]] = Field(
        serialization_alias="natPortForwarding", validation_alias="natPortForwarding", default=None
    )
    static_nat: Optional[List[StaticNat]] = Field(
        serialization_alias="staticNat", validation_alias="staticNat", default=None
    )
    static_nat_subnet: Optional[List[StaticNatSubnet]] = Field(
        serialization_alias="staticNatSubnet", validation_alias="staticNatSubnet", default=None
    )
    nat64_v4_pool: Optional[List[Nat64v4Pool]] = Field(
        serialization_alias="nat64V4Pool", validation_alias="nat64V4Pool", default=None
    )
    route_leak_from_global: Optional[List[RouteLeakFromGlobal]] = Field(
        serialization_alias="routeLeakFromGlobal", validation_alias="routeLeakFromGlobal", default=None
    )
    route_leak_from_service: Optional[List[RouteLeakFromService]] = Field(
        serialization_alias="routeLeakFromService", validation_alias="routeLeakFromService", default=None
    )
    route_leak_between_services: Optional[List[RouteLeakBetweenServices]] = Field(
        serialization_alias="routeLeakBetweenServices", validation_alias="routeLeakBetweenServices", default=None
    )
    mpls_vpn_ipv4_route_target: Optional[MplsVpnIPv4RouteTarget] = Field(
        serialization_alias="mplsVpnIpv4RouteTarget", validation_alias="mplsVpnIpv4RouteTarget", default=None
    )
    mpls_vpn_ipv6_route_target: Optional[MplsVpnIPv6RouteTarget] = Field(
        serialization_alias="mplsVpnIpv6RouteTarget", validation_alias="mplsVpnIpv6RouteTarget", default=None
    )
    enable_sdra: Optional[Union[Global[bool], Default[bool]]] = Field(
        serialization_alias="enableSdra", validation_alias="enableSdra", default=None
    )


class LanVpnCreationPayload(BaseModel):
    name: str
    description: Optional[str] = None
    data: LanVpnData
    metadata: Optional[dict] = None



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from catalystwan.models.policy.policy_definition import PolicyDefinitionBase

QoSScheduling = Literal[
    "llq",
    "wrr",
]

QoSDropType = Literal[
    "tail-drop",
    "red-drop",
]


class QoSScheduler(BaseModel):
    queue: str
    class_map_ref: Union[UUID, Literal[""]] = Field(serialization_alias="classMapRef", validation_alias="classMapRef")
    bandwidth_percent: str = Field("1", serialization_alias="bandwidthPercent", validation_alias="bandwidthPercent")
    buffer_percent: str = Field("1", serialization_alias="bufferPercent", validation_alias="bufferPercent")
    burst: Optional[str] = None
    scheduling: QoSScheduling = "wrr"
    drops: QoSDropType = "tail-drop"
    temp_key_values: Optional[str] = Field(
        default=None, serialization_alias="tempKeyValues", validation_alias="tempKeyValues"
    )

    @staticmethod
    def get_default_control_scheduler() -> "QoSScheduler":
        return QoSScheduler(
            queue="0",
            class_map_ref="",
            bandwidth_percent="100",
            buffer_percent="100",
            burst="15000",
            scheduling="llq",
            drops="tail-drop",
        )

    model_config = ConfigDict(populate_by_name=True)

    @field_validator("queue")
    @classmethod
    def check_queue(cls, queue_str: str):
        assert 0 <= int(queue_str) <= 7
        return queue_str

    @field_validator("bandwidth_percent", "buffer_percent")
    @classmethod
    def check_bandwidth_and_buffer_percent(cls, percent_str: str):
        assert 1 <= int(percent_str) <= 100
        return percent_str

    @field_validator("burst")
    @classmethod
    def check_burst(cls, burst_val: Union[str, None]):
        if burst_val is not None:
            assert 5000 <= int(burst_val) <= 10_000_000
        return burst_val


class QoSMapDefinition(BaseModel):
    qos_schedulers: List[QoSScheduler] = Field(serialization_alias="qosSchedulers", validation_alias="qosSchedulers")
    model_config = ConfigDict(populate_by_name=True)


class QoSMapPolicy(PolicyDefinitionBase):
    type: Literal["qosMap"] = "qosMap"
    definition: QoSMapDefinition = QoSMapDefinition(qos_schedulers=[])
    model_config = ConfigDict(populate_by_name=True)

    def add_scheduler(
        self,
        queue: int,
        class_map_ref: UUID,
        bandwidth: int = 1,
        buffer: int = 1,
        scheduling: QoSScheduling = "wrr",
        drops: QoSDropType = "tail-drop",
        burst: Optional[int] = None,
    ) -> None:
        self.definition.qos_schedulers.append(
            QoSScheduler(
                queue=str(queue),
                class_map_ref=class_map_ref,
                bandwidth_percent=str(bandwidth),
                buffer_percent=str(buffer),
                burst=str(burst) if burst is not None else None,
                scheduling=scheduling,
                drops=drops,
            )
        )

    @model_validator(mode="after")
    def generate_default_control_scheduler(self):
        if not self.definition.qos_schedulers:
            # Only when creating (not when value obtained from remote is present)
            self.definition = QoSMapDefinition(qos_schedulers=[QoSScheduler.get_default_control_scheduler()])
        return self



# Copyright 2022 Cisco Systems, Inc. and its affiliates

import datetime as dt
from typing import Any, ClassVar, Dict, Iterable, List, Protocol, Type, TypeVar, runtime_checkable

import attrs  # type: ignore
from attr import Attribute, fields, fields_dict
from dateutil import parser  # type: ignore

T = TypeVar("T")
FIELD_NAME = "__field_name"


@runtime_checkable
class AttrsInstance(Protocol):
    __attrs_attrs__: ClassVar[Any]


def create_dataclass(cls: Type[T], data: Dict[str, Any]) -> T:
    """Deserializes data to convert it into an object.

    Prepares data to create dataclass from json structure.
    If field has given metadata information about FIELD_NAME,
    the data keys are mapped into an field name from given dataclass.
    Then it filters fields which are not given explicitly.
    Creation is handled by attrs module.

    Args:
        cls (type): Dataclass which should be created
        data (dict): A dict to deserialize from

    Returns:
        A dataclass with implemented fields
    """

    def filter_fields(available_fields: Iterable[str], data: Dict[str, Any]) -> Dict[str, Any]:
        return dict(filter(lambda key_value: key_value[0] in available_fields, data.items()))

    data_copy = data.copy()
    for field in fields(cls):  # type: ignore[misc]
        json_field_name = field.metadata.get(FIELD_NAME, None)
        if json_field_name and json_field_name in data_copy:
            data_copy[field.name] = data_copy.pop(json_field_name)
    filtered_data = filter_fields(fields_dict(cls).keys(), data_copy)
    return cls(**filtered_data)


def convert_attributes(cls: type, fields: List[Attribute]) -> List[Attribute]:
    """Automatically add converters to the attributes based on their types.

    Args:
        cls (type): Class right before being converted into an attrs class
        fields (List[Attribute]): List of all Attribute instances that
            will later be set to the __attrs_attrs__

    Returns:
        A modified list of all attr.Attribute instances with proper converters
    """
    results = []
    for field in fields:
        if field.converter is not None:
            results.append(field)
            continue
        if field.type in {dt.datetime, "datetime"}:
            converter = (
                lambda d: parser.parse(d) if isinstance(d, str) else dt.datetime.fromtimestamp(d / 1000)
            )  # type: ignore # noqa: E731
        elif field.type in {str, "str"}:
            converter = lambda x: str(x)  # type: ignore # noqa: E731
        else:
            converter = None
        results.append(field.evolve(converter=converter))  # type: ignore
    return results


def asdict(dataclass: AttrsInstance) -> dict:
    """Serializes attrs dataclass into a python dictionary.

    While serializing dataclass, FIELD_NAME is taken into account.

    Args:
        dataclass (Type[T]): Valid attrs dataclass

    Returns:
        dict: Serialized dict
    """

    json_fields_excluded = attrs.asdict(dataclass, filter=lambda x, _: FIELD_NAME not in x.metadata)
    json_fields = attrs.asdict(dataclass, filter=lambda x, _: FIELD_NAME in x.metadata)

    for field in fields(dataclass.__class__):  # type: ignore[misc]
        field_value = getattr(dataclass, field.name)

        json_field_name = field.metadata.get(FIELD_NAME, None)
        if json_field_name:
            json_fields[json_field_name] = json_fields.pop(field.name)
            if isinstance(field_value, AttrsInstance):
                json_fields[json_field_name] = asdict(field_value)

            if isinstance(field_value, list):  # tuple, List[AttrsInstance], set, frozenset
                if len(field_value) > 0 and isinstance(field_value[0], AttrsInstance):
                    json_fields[json_field_name] = [asdict(_f) for _f in field_value]
        else:
            if isinstance(field_value, AttrsInstance):
                json_fields_excluded[field.name] = asdict(field_value)

    return {**json_fields, **json_fields_excluded}


def flatten_dict(d: Dict) -> Dict:
    """Flattens the given dictionary.

    If it has nested dicts or lists of dicts, inlines them into
    the original dictionary. Other member types are ignored.
    For nested lists only members of type dict is processed.

    Args:
        d (Dict): Dictionary to flatten

    Returns:
        A flattened dictionary
    """
    if not isinstance(d, dict):
        return {}

    def recurse(key, value):
        if isinstance(value, dict):
            for k, v in value.items():
                yield from recurse(k, v)
        elif isinstance(value, list):
            for v in value:
                yield from recurse(None, v)
        elif key is not None:
            yield key, value

    return dict(recurse(None, d))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from typing import Any, List

from pydantic import BaseModel, ConfigDict, Field


class FeatureTemplatePayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    name: str = Field(alias="templateName")
    description: str = Field(alias="templateDescription")
    template_type: str = Field(alias="templateType")  # Enum
    device_types: List[str] = Field(alias="deviceType")  # Enum
    default: bool = Field(alias="factoryDefault", default=False)
    version: str = Field(alias="templateMinVersion", default="15.0.0")  # Enum
    definition: Any = Field(alias="templateDefinition")



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from uuid import UUID

from catalystwan.endpoints import APIEndpoints, delete, get, post, put
from catalystwan.models.policy.definitions.zone_based_firewall import ZoneBasedFWPolicy
from catalystwan.models.policy.policy_definition import (
    PolicyDefinitionEditResponse,
    PolicyDefinitionEndpoints,
    PolicyDefinitionGetResponse,
    PolicyDefinitionId,
    PolicyDefinitionInfo,
    PolicyDefinitionPreview,
)
from catalystwan.typed_list import DataSequence


class ZoneBasedFWPolicyEditPayload(ZoneBasedFWPolicy, PolicyDefinitionId):
    pass


class ZoneBasedFWPolicyGetResponse(ZoneBasedFWPolicy, PolicyDefinitionGetResponse):
    pass


class ConfigurationPolicyZoneBasedFirewallDefinition(APIEndpoints, PolicyDefinitionEndpoints):
    @post("/template/policy/definition/zonebasedfw")
    def create_policy_definition(self, payload: ZoneBasedFWPolicy) -> PolicyDefinitionId:
        ...

    @delete("/template/policy/definition/zonebasedfw/{id}")
    def delete_policy_definition(self, id: UUID) -> None:
        ...

    def edit_multiple_policy_definition(self):
        # PUT /template/policy/definition/zonebasedfw/multiple/{id}
        ...

    @put("/template/policy/definition/zonebasedfw/{id}")
    def edit_policy_definition(self, id: UUID, payload: ZoneBasedFWPolicyEditPayload) -> PolicyDefinitionEditResponse:
        ...

    @get("/template/policy/definition/zonebasedfw", "data")
    def get_definitions(self) -> DataSequence[PolicyDefinitionInfo]:
        ...

    @get("/template/policy/definition/zonebasedfw/{id}")
    def get_policy_definition(self, id: UUID) -> ZoneBasedFWPolicyGetResponse:
        ...

    @post("/template/policy/definition/zonebasedfw/preview")
    def preview_policy_definition(self, payload: ZoneBasedFWPolicy) -> PolicyDefinitionPreview:
        ...

    @get("/template/policy/definition/zonebasedfw/preview/{id}")
    def preview_policy_definition_by_id(self, id: UUID) -> PolicyDefinitionPreview:
        ...

    def save_policy_definition_in_bulk(self):
        # PUT /template/policy/definition/zonebasedfw/bulk
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# type: ignore
import unittest
from unittest import TestCase

from parameterized import parameterized

from catalystwan.api.templates.feature_template_field import get_path_dict


class TestGetPathDict(TestCase):
    @parameterized.expand(
        [
            ([], {}),
            (["a"], {"a": {}}),
            (["a", "b"], {"a": {}, "b": {}}),
            ([["a", "b"]], {"a": {"b": {}}}),
            ([["a", "b"], ["c"]], {"a": {"b": {}}, "c": {}}),
            ([["a", "b"], ["c"]], {"a": {"b": {}}, "c": {}}),
            ([["a", "b"], ["c"], ["a", "d"]], {"a": {"b": {}, "d": {}}, "c": {}}),
        ]
    )
    def test_get_path_dict(self, paths, answer):
        # Arrange, Act
        output = get_path_dict(paths)

        # Assert
        self.assertEqual(output, answer)


if __name__ == "__main__":
    unittest.main()



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class FQDNListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    pattern: Global[str] = Field(
        description="Ex: cisco.com, .*cisco.com, .*.cisco.com. Should not start with '*' or '+'"
    )


class FQDNDomainParcel(_ParcelBase):
    entries: List[FQDNListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def from_fqdns(self, fqdns: List[str]):
        for fqdn in fqdns:
            self.entries.append(FQDNListEntry(pattern=as_global(fqdn)))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, delete, get, post, put, versions
from catalystwan.models.configuration.common import Solution
from catalystwan.models.configuration.feature_profile.common import ProfileType
from catalystwan.typed_list import DataSequence


class ProfileId(BaseModel):
    id: str


# TODO Get mode from schema
class ConfigGroupCreationPayload(BaseModel):
    name: str
    description: str
    solution: Solution
    profiles: Optional[List[ProfileId]]


class FeatureProfile(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    id: str
    name: str
    description: Optional[str]
    solution: str
    type: ProfileType
    created_by: str = Field(serialization_alias="createdBy", validation_alias="createdBy")
    last_updated_by: str = Field(serialization_alias="lastUpdatedBy", validation_alias="lastUpdatedBy")
    created_on: datetime = Field(serialization_alias="createdOn", validation_alias="createdOn")
    last_updated_on: datetime = Field(serialization_alias="lastUpdatedOn", validation_alias="lastUpdatedOn")


class ConfigGroup(BaseModel):
    name: str
    description: Optional[str]
    solution: Solution
    profiles: Optional[List[FeatureProfile]]


class ConfigGroupResponsePayload(BaseModel):
    config_groups: List[ConfigGroup]


class ConfigGroupEditPayload(BaseModel):
    name: str
    description: str
    solution: Solution
    profiles: Optional[List[ProfileId]]


class DeviceId(BaseModel):
    id: str


class ConfigGroupAssociatePayload(BaseModel):
    devices: List[DeviceId]


class ConfigGroupVariablesCreatePayload(BaseModel):
    deviceIds: List[str]
    suggestions: bool = True


class VariableData(BaseModel):
    name: str
    value: str


class DeviceVariables(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    device_id: str = Field(serialization_alias="device-id", validation_alias="device-id")
    variables: List[VariableData]


class GroupVariables(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    name: str
    group_variables: List[VariableData] = Field(
        serialization_alias="group-variables", validation_alias="group-variables"
    )


class ConfigGroupVariablesCreateResponse(BaseModel):
    family: Solution
    devices: List[DeviceVariables]
    groups: List[GroupVariables]


class ConfigGroupVariablesEditPayload(BaseModel):
    solution: Solution
    devices: List[DeviceVariables]


class ConfigGroupDeployPayload(BaseModel):
    devices: List[DeviceId]


class ConfigGroupDeployResponse(BaseModel):
    parentTaskId: str


class ConfigGroupDisassociateResponse(BaseModel):
    parentTaskId: str


class ConfigGroupCreationResponse(BaseModel):
    id: str


class EditedProfileId(BaseModel):
    profileId: str


class ConfigGroupEditResponse(BaseModel):
    id: str
    profiles: List[EditedProfileId]


class ConfigurationGroup(APIEndpoints):
    @versions(supported_versions=(">=20.9"), raises=False)
    @put("/v1/config-group/{config_group_id}/device/associate")
    def associate(self, config_group_id: str, payload: ConfigGroupAssociatePayload) -> None:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/config-group")
    def create_config_group(self, payload: ConfigGroupCreationPayload) -> ConfigGroupCreationResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/config-group/{config_group_id}/device/variables")
    def create_variables(
        self, config_group_id: str, payload: ConfigGroupVariablesCreatePayload
    ) -> ConfigGroupVariablesCreateResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @delete("/v1/config-group/{config_group_id}")
    def delete_config_group(self, config_group_id: str) -> None:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/config-group/{config_group_id}/device/deploy")
    def deploy(self, config_group_id: str, payload: ConfigGroupDeployPayload) -> ConfigGroupDeployResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @delete("/v1/config-group/{config_group_id}/device/associate")
    def disassociate(
        self, config_group_id: str, payload: ConfigGroupAssociatePayload
    ) -> ConfigGroupDisassociateResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @put("/v1/config-group/{config_group_id}")
    def edit_config_group(self, config_group_id: str, payload: ConfigGroupEditPayload) -> ConfigGroupEditResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/config-group")
    def get(self) -> DataSequence[ConfigGroup]:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @put("/v1/config-group/{config_group_id}/device/variables")
    def update_variables(self, config_group_id: str, payload: ConfigGroupVariablesEditPayload) -> None:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import Optional

from catalystwan.api.configuration_groups.parcel import _ParcelBase
from catalystwan.endpoints import JSON, APIEndpoints, delete, get, post, put, versions
from catalystwan.models.configuration.feature_profile.common import (
    FeatureProfileCreationPayload,
    FeatureProfileCreationResponse,
    FeatureProfileInfo,
    GetFeatureProfilesPayload,
    ParcelId,
    SchemaTypeQuery,
)
from catalystwan.typed_list import DataSequence


class SystemFeatureProfile(APIEndpoints):
    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan/system/aaa/schema", resp_json_key="request")
    def get_sdwan_system_aaa_parcel_schema(self, params: SchemaTypeQuery) -> JSON:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan/system")
    def get_sdwan_system_feature_profiles(
        self, payload: Optional[GetFeatureProfilesPayload]
    ) -> DataSequence[FeatureProfileInfo]:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/feature-profile/sdwan/system")
    def create_sdwan_system_feature_profile(
        self, payload: FeatureProfileCreationPayload
    ) -> FeatureProfileCreationResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @delete("/v1/feature-profile/sdwan/system/{system_id}")
    def delete_sdwan_system_feature_profile(self, system_id: str) -> None:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/feature-profile/sdwan/system/{system_id}/aaa")
    def create_aaa_profile_parcel_for_system(self, system_id: str, payload: _ParcelBase) -> ParcelId:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @put("/v1/feature-profile/sdwan/system/{system_id}/aaa/{parcel_id}")
    def edit_aaa_profile_parcel_for_system(self, system_id: str, parcel_id: str, payload: _ParcelBase) -> ParcelId:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv6Network
from typing import Any, List, Literal, Optional, Set, Tuple, Union
from uuid import UUID

from pydantic import ConfigDict, Field
from typing_extensions import Annotated

from catalystwan.models.policy.policy_definition import (
    CountAction,
    DefaultAction,
    DefinitionWithSequencesCommonBase,
    DestinationDataIPv6PrefixListEntry,
    DestinationIPv6Entry,
    DestinationPortEntry,
    DeviceAccessProtocol,
    Match,
    PolicyActionType,
    PolicyDefinitionBase,
    PolicyDefinitionSequenceBase,
    SourceDataIPv6PrefixListEntry,
    SourceIPv6Entry,
    SourcePortEntry,
)

DeviceAccessIPv6PolicySequenceMatchEntry = Annotated[
    Union[
        DestinationDataIPv6PrefixListEntry,
        DestinationIPv6Entry,
        DestinationPortEntry,
        SourceDataIPv6PrefixListEntry,
        SourceIPv6Entry,
        SourcePortEntry,
    ],
    Field(discriminator="field"),
]

DeviceAccessIPv6PolicySequenceActions = Any  # TODO


class DeviceAccessIPv6PolicyHeader(PolicyDefinitionBase):
    type: Literal["deviceAccessPolicyv6"] = "deviceAccessPolicyv6"


class DeviceAccessIPv6PolicySequenceMatch(Match):
    entries: List[DeviceAccessIPv6PolicySequenceMatchEntry] = []


class DeviceAccessIPv6PolicySequence(PolicyDefinitionSequenceBase):
    sequence_type: Literal["deviceaccesspolicyv6"] = Field(
        default="deviceaccesspolicyv6", serialization_alias="sequenceType", validation_alias="sequenceType"
    )
    base_action: PolicyActionType = Field(
        default="accept", serialization_alias="baseAction", validation_alias="baseAction"
    )
    match: DeviceAccessIPv6PolicySequenceMatch = DeviceAccessIPv6PolicySequenceMatch()
    actions: List[DeviceAccessIPv6PolicySequenceActions] = []
    model_config = ConfigDict(populate_by_name=True)

    def match_device_access_protocol(self, port: DeviceAccessProtocol) -> None:
        self._insert_match(DestinationPortEntry.from_port_set_and_ranges(ports={port}))

    def match_source_data_prefix_list(self, data_prefix_list_id: UUID) -> None:
        self._insert_match(SourceDataIPv6PrefixListEntry(ref=data_prefix_list_id))

    def match_source_ip(self, networks: List[IPv6Network]) -> None:
        self._insert_match(SourceIPv6Entry.from_ipv6_networks(networks))

    def match_source_port(self, ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> None:
        self._insert_match(SourcePortEntry.from_port_set_and_ranges(ports, port_ranges))

    def match_destination_data_prefix_list(self, data_prefix_list_id: UUID) -> None:
        self._insert_match(DestinationDataIPv6PrefixListEntry(ref=data_prefix_list_id))

    def match_destination_ip(self, networks: List[IPv6Network]) -> None:
        self._insert_match(DestinationIPv6Entry.from_ipv6_networks(networks))

    def associate_count_action(self, counter_name: str) -> None:
        self._insert_action(CountAction(parameter=counter_name))


class DeviceAccessIPv6Policy(DeviceAccessIPv6PolicyHeader, DefinitionWithSequencesCommonBase):
    sequences: List[DeviceAccessIPv6PolicySequence] = []
    default_action: DefaultAction = Field(
        default=DefaultAction(type="drop"),
        serialization_alias="defaultAction",
        validation_alias="defaultAction",
    )
    model_config = ConfigDict(populate_by_name=True)

    def add_acl_sequence(
        self,
        name: str = "Device Access Control List",
        base_action: PolicyActionType = "accept",
        device_access_protocol: Optional[DeviceAccessProtocol] = None,
    ) -> DeviceAccessIPv6PolicySequence:
        seq = DeviceAccessIPv6PolicySequence(
            sequence_name=name,
            base_action=base_action,
            sequence_ip_type="ipv4",
        )
        if device_access_protocol is not None:
            seq.match_device_access_protocol(port=device_access_protocol)
        self.add(seq)
        return seq



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest

from parameterized import parameterized  # type: ignore

from catalystwan.api.templates.models.cisco_aaa_model import CiscoAAAModel
from catalystwan.api.templates.models.cisco_system import CiscoSystemModel
from catalystwan.api.templates.models.omp_vsmart_model import OMPvSmart
from catalystwan.api.templates.models.security_vsmart_model import SecurityvSmart
from catalystwan.api.templates.models.system_vsmart_model import SystemVsmart
from catalystwan.utils.feature_template.choose_model import choose_model


class TestChooseModel(unittest.TestCase):
    @parameterized.expand(
        [
            ("cedge_aaa", CiscoAAAModel),
            ("omp-vsmart", OMPvSmart),
            ("security-vsmart", SecurityvSmart),
            ("system-vsmart", SystemVsmart),
            ("cisco_system", CiscoSystemModel),
        ]
    )
    def test_choose_model(self, model_type, model_cls):
        # Arrange
        name = "My model name"
        description = "My model description"
        model_from_cls = model_cls(template_name=name, template_description=description)

        # Act
        chosen_model = choose_model(model_type)
        model_from_choice = chosen_model(template_name=name, template_description=description)

        # Assert
        self.assertEqual(model_from_choice, model_from_cls)



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from unittest import TestCase
from unittest.mock import patch

from parameterized import parameterized  # type: ignore
from pytest import mark  # type: ignore
from tenacity import RetryError  # type: ignore

from catalystwan.api.basic_api import DevicesAPI, DeviceStateAPI
from catalystwan.dataclasses import BfdSessionData, Connection, Device, WanInterface
from catalystwan.endpoints.endpoints_container import APIEndpointContainter
from catalystwan.endpoints.monitoring.device_details import DeviceData
from catalystwan.endpoints.real_time_monitoring.reboot_history import RebootEntry
from catalystwan.exceptions import CatalystwanException
from catalystwan.response import ManagerResponse
from catalystwan.typed_list import DataSequence
from catalystwan.utils.creation_tools import create_dataclass
from catalystwan.utils.personality import Personality


class ResponseMock:
    def __init__(self, json, headers={}, cookies={}):
        self._json = json
        self.headers = headers
        self.cookies = cookies

    def json(self):
        return self._json


class TestDevicesAPI(TestCase):
    def setUp(self) -> None:
        self.devices = [
            {  # vmanage
                "device-model": "vmanage",
                "deviceId": "1.1.1.1",
                "uuid": "aaaaaaaa-6169-445c-8e49-c0bdaaaaaaa",
                "cpuLoad": 4.71,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.1",
                "board-serial": "11122233",
                "personality": "vmanage",
                "memUsage": 57.0,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm200",
                "cpuState": "normal",
                "chassis-number": "aaaaaaaa-6aa9-445c-8e49-c0aaaaaaaaa9",
            },
            {  # vsmart
                "device-model": "vsmart",
                "deviceId": "1.1.1.3",
                "uuid": "bbcccccc-6169-445c-8e49-c0bdccccccc",
                "cpuLoad": 2.76,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.3",
                "board-serial": "11223399",
                "personality": "vsmart",
                "memUsage": 28.0,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm129",
                "cpuState": "normal",
                "chassis-number": "abbbbbbb-6169-445c-8e49-c0bccccccccc",
            },
            {  # vbond
                "device-model": "vedge-cloud",
                "deviceId": "1.1.1.2",
                "uuid": "bbbbbbbb-6169-445c-8e49-c0bdaaaaaaa",
                "cpuLoad": 1.76,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.2",
                "board-serial": "11223344",
                "personality": "vbond",
                "memUsage": 26.0,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm128",
                "cpuState": "normal",
                "chassis-number": "abbbbbbb-6169-445c-8e49-c0bbbbbbbbb9",
            },
            {  # vedge
                "device-model": "vedge-cloud",
                "deviceId": "169.254.10.10",
                "uuid": "bc2a78ac-a06e-40fd-b2b7-1b1e062f3f9e",
                "cpuLoad": 1.46,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "172.16.254.2",
                "board-serial": "12345708",
                "personality": "vedge",
                "memUsage": 24.1,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm1",
                "cpuState": "normal",
                "chassis-number": "bc2a78ac-a06e-40fd-b2b7-1b1e062f3f9e",
            },
        ]
        self.devices_dataclass = [create_dataclass(Device, device) for device in self.devices]
        self.devices_dataseq = DataSequence(Device, [create_dataclass(Device, device) for device in self.devices])
        self.controllers_dataclass = DataSequence(
            Device, [create_dataclass(Device, device) for device in self.devices[:2]]
        )
        self.orchestrators_dataseq = DataSequence(Device, [create_dataclass(Device, self.devices[2])])
        self.vsmarts_dataseq = DataSequence(Device, [create_dataclass(Device, self.devices[1])])
        self.vbonds_dataseq = DataSequence(Device, [create_dataclass(Device, self.devices[2])])
        self.edges_dataseq = DataSequence(Device, [create_dataclass(Device, self.devices[3])])
        self.system_ips_list = [device["local-system-ip"] for device in self.devices]
        self.ips_list = [device["deviceId"] for device in self.devices]
        self.list_all_devices_resp = DataSequence(DeviceData, [DeviceData.model_validate(dev) for dev in self.devices])

    @patch.object(DevicesAPI, "get")
    def test_controllers(self, mock_devices):
        # # Arrange
        # MockDevices = Mock()
        # mock_devices.return_value = MockDevices
        # session = Mock()
        # test_object = DevicesAPI(session)
        # test_object.devices = self.devices_dataclass
        # # Act
        # answer = test_object.get().filter(personality=[Personality.VMANAGE, Personality.VSMART])
        # # Assert
        # self.assertEqual(answer, self.controllers_dataclass)
        pass  # TODO fix after updating .filter()

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_orchestrators(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).get().filter(personality=Personality.VBOND)

        # Assert
        self.assertEqual(answer, self.orchestrators_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_edges(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).get().filter(personality=Personality.EDGE)

        # Assert
        self.assertEqual(answer, self.edges_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_vsmarts(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).get().filter(personality=Personality.VSMART)

        # Assert
        self.assertEqual(answer, self.vsmarts_dataseq)

    @patch("catalystwan.session.ManagerSession")
    @patch.object(DevicesAPI, "get")
    def test_system_ips(self, mock_devices, mock_session):
        # Arrange
        mock_devices.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).system_ips

        # Assert
        self.assertEqual(answer, self.system_ips_list)

    @patch("catalystwan.session.ManagerSession")
    @patch.object(DevicesAPI, "get")
    def test_ips(self, mock_devices, mock_session):
        # Arrange
        mock_devices.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).ips

        # Assert
        self.assertEqual(answer, self.ips_list)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).get()

        # Assert
        self.assertEqual(answer, self.devices_dataseq)

    @parameterized.expand(
        [
            ["aaaaaaaa-6169-445c-8e49-c0bdaaaaaaa", 0],
            ["bbcccccc-6169-445c-8e49-c0bdccccccc", 1],
            ["bbbbbbbb-6169-445c-8e49-c0bdaaaaaaa", 2],
            ["bc2a78ac-a06e-40fd-b2b7-1b1e062f3f9e", 3],
        ]
    )
    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_device_details(self, uuid, device_number, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = DataSequence(
            Device, [create_dataclass(Device, self.devices[device_number])]
        )

        # Act
        answer = DevicesAPI(mock_session).get_device_details(uuid=uuid)

        # Assert
        self.assertEqual(answer, create_dataclass(Device, self.devices[device_number]))

    @parameterized.expand(
        [[Personality.EDGE, 1], [Personality.VBOND, 1], [Personality.VMANAGE, 1], [Personality.VSMART, 1]]
    )
    @patch("catalystwan.session.ManagerSession")
    @patch.object(DevicesAPI, "get")
    def test_count_devices(self, personality, devices_number, mock_get, mock_session):
        # Arrange
        mock_get.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).count_devices(personality=personality)

        # Assert
        self.assertEqual(answer, devices_number)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_reachable_devices_vsmarts(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = DataSequence(Device, [create_dataclass(Device, self.devices[1])])

        # Act
        answer = DevicesAPI(mock_session).get_reachable_devices(personality=Personality.VSMART)

        # Assert
        self.assertEqual(answer, self.vsmarts_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_reachable_devices_vbonds(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = DataSequence(Device, [create_dataclass(Device, self.devices[2])])

        # Act
        answer = DevicesAPI(mock_session).get_reachable_devices(personality=Personality.VBOND)

        # Assert
        self.assertEqual(answer, self.vbonds_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_reachable_devices_vedgess(self, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = DataSequence(Device, [create_dataclass(Device, self.devices[3])])

        # Act
        answer = DevicesAPI(mock_session).get_reachable_devices(personality=Personality.EDGE)

        # Assert
        self.assertEqual(answer, self.edges_dataseq)

    @patch("catalystwan.session.ManagerSession")
    def test_get_reachable_devices_for_vmanage(self, mock_session, personality=Personality.VMANAGE):
        # Arrange
        mock_session.get_data.return_value = [
            device for device in self.devices if device["personality"] == personality.value
        ]

        # Act
        def answer():
            return DevicesAPI(mock_session).get_reachable_devices(personality)

        # Assert
        self.assertRaises(AssertionError, answer)

    @patch("catalystwan.session.ManagerSession")
    def test_send_certificate_state_to_controllers(self, mock_session):
        # Arrange
        mock_session.post().json.return_value = {"id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}

        # Act
        answer = DevicesAPI(mock_session).send_certificate_state_to_controllers()

        # Assert
        self.assertTrue(answer)

    @patch("catalystwan.session.ManagerSession")
    def test_send_certificate_state_to_controllers_error(self, mock_session):
        # Arrange
        mock_session.post().json.return_value = {}

        # Act
        def answer():
            return DevicesAPI(mock_session).send_certificate_state_to_controllers()

        # Assert
        self.assertRaises(CatalystwanException, answer)

    @parameterized.expand([["vm200", 0], ["vm129", 1], ["vm128", 2], ["vm1", 3]])
    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_by_hostname(self, hostname, device_number, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).get().filter(hostname=hostname)

        # Assert
        self.assertEqual(answer, DataSequence(Device, [create_dataclass(Device, self.devices[device_number])]))

    @parameterized.expand([["1.1.1.1", 0], ["1.1.1.3", 1], ["1.1.1.2", 2], ["169.254.10.10", 3]])
    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_by_id(self, device_id, device_number, mock_session, mock_response):
        # Arrange
        mock_session.get.return_value = mock_response
        mock_session.endpoints.monitoring_device_details.list_all_devices.return_value = self.list_all_devices_resp
        mock_response.dataseq.return_value = self.devices_dataseq

        # Act
        answer = DevicesAPI(mock_session).get().filter(id=device_id)

        # Assert
        self.assertEqual(answer, DataSequence(Device, [create_dataclass(Device, self.devices[device_number])]))


class TestDevicesStateAPI(TestCase):
    def setUp(self) -> None:
        self.crash_info = [
            {
                "core-time": "Tue Feb 15 04:14:38 UTC 2022",
                "vdevice-dataKey": "169.254.10.12-0",
                "vdevice-name": "169.254.10.12",
                "index": 0,
                "lastupdated": 1645064726542,
                "core-filename": "vm5_htx_19219_20220215-041430-UTC.core.gz",
                "core-time-date": 1644898478000,
                "vdevice-host-name": "vm5",
            },
            {
                "core-time": "Tue Feb 15 04:15:24 UTC 2022",
                "vdevice-dataKey": "169.254.10.12-1",
                "vdevice-name": "169.254.10.12",
                "index": 1,
                "lastupdated": 1645064726542,
                "core-filename": "vm5_htx_19539_20220215-041515-UTC.core.gz",
                "core-time-date": 1644898524000,
                "vdevice-host-name": "vm5",
            },
        ]
        self.connections_info = [
            {
                "system-ip": "1.1.1.1",
                "peer-type": "vsmart",
                "state": "up",
            },
            {
                "system-ip": "1.1.1.2",
                "peer-type": "vedge",
                "state": "up",
            },
        ]
        self.connections_info_dataclass = [create_dataclass(Connection, item) for item in self.connections_info]
        self.reboot_history = [
            {
                "reboot_date_time-date": 1642651714000,
                "lastupdated": 1642656685355,
                "vdevice-dataKey": "172.16.255.11-2022-01-20T04:08:34+00:00",
                "reboot_date_time": "2022-01-20T04:08:34+00:00",
                "reboot_reason": "Initiated by user - Reboot issued via NETCONF",
                "vdevice-host-name": "vm1",
                "vdevice-name": "172.16.255.11",
            },
            {
                "reboot_date_time-date": 164265171400,
                "lastupdated": 1642656685356,
                "vdevice-dataKey": "172.16.255.11-2022-01-20T04:08:34+00:00",
                "reboot_date_time": "2022-01-20T04:08:34+00:00",
                "reboot_reason": "Initiated by user - Reboot issued via NETCONF",
                "vdevice-host-name": "vm2",
                "vdevice-name": "172.16.255.12",
            },
        ]
        self.reboot_history_payload = DataSequence(
            RebootEntry, [RebootEntry.model_validate(item) for item in self.reboot_history]
        )
        self.device = [
            {  # vmanage
                "device-model": "vmanage",
                "deviceId": "1.1.1.1",
                "uuid": "aaaaaaaa-6169-445c-8e49-c0bdaaaaaaa",
                "cpuLoad": 4.71,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.1",
                "board-serial": "11122233",
                "personality": "vmanage",
                "memUsage": 57.0,
                "reachability": "reachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm200",
                "cpuState": "normal",
                "chassis-number": "aaaaaaaa-6aa9-445c-8e49-c0aaaaaaaaa9",
            }
        ]
        self.device_dataclass = create_dataclass(Device, self.device[0])
        self.wan_interfaces = [
            {
                "color": "default",
                "vdevice-name": "1.1.1.1",
                "admin-state": "up",
                "interface": "eth1",
                "private-port": 12345,
                "vdevice-host-name": "vm1",
                "public-ip": "1.1.1.1",
                "operation-state": "up",
                "public-port": 12345,
                "private-ip": "1.1.1.1",
            },
            {
                "color": "default",
                "vdevice-name": "1.1.1.1",
                "admin-state": "up",
                "interface": "eth1",
                "private-port": 11111,
                "vdevice-host-name": "vm2",
                "public-ip": "1.1.1.1",
                "operation-state": "up",
                "public-port": 11111,
                "private-ip": "1.1.1.1",
            },
        ]
        self.wan_interfaces_dataclass = [create_dataclass(WanInterface, item) for item in self.wan_interfaces]
        self.bfd_session = [
            {
                "src-ip": "1.1.1.1",
                "dst-ip": "1.1.1.2",
                "color": "3g",
                "system-ip": "1.1.1.1",
                "site-id": 1,
                "local-color": "3g",
                "state": "up",
            },
            {
                "src-ip": "1.1.1.1",
                "dst-ip": "1.1.1.3",
                "color": "3g",
                "system-ip": "1.1.1.1",
                "site-id": 1,
                "local-color": "3g",
                "state": "up",
            },
        ]
        self.bfd_session_dataclass = [create_dataclass(BfdSessionData, item) for item in self.bfd_session]
        self.bfd_session_down = [
            {
                "src-ip": "1.1.1.1",
                "dst-ip": "1.1.1.2",
                "color": "3g",
                "system-ip": "1.1.1.1",
                "site-id": 1,
                "local-color": "3g",
                "state": "down",
            }
        ]
        self.device_unreachable = [
            {  # vmanage
                "device-model": "vmanage",
                "deviceId": "1.1.1.1",
                "uuid": "zzzzzzzz-6169-445c-8e49-c0bdaaaaaaa",
                "cpuLoad": 4.71,
                "state_description": "All daemons up",
                "status": "normal",
                "memState": "normal",
                "local-system-ip": "1.1.1.1",
                "board-serial": "11122233",
                "personality": "vmanage",
                "memUsage": 57.0,
                "reachability": "unreachable",
                "connectedVManages": ["1.1.1.1"],
                "host-name": "vm200",
                "cpuState": "normal",
                "chassis-number": "aaaaaaaa-6aa9-445c-8e49-c0aaaaaaaaa9",
            }
        ]

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_crash_info(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.crash_info
        # Act
        answer = DeviceStateAPI(mock_session).get_device_crash_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.crash_info)

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_crash_info_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_crash_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_control_connections_info(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.connections_info
        # Act
        answer = DeviceStateAPI(mock_session).get_device_control_connections_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.connections_info_dataclass)

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_control_connections_info_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_control_connections_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_orchestrator_connections_info(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.connections_info
        # Act
        answer = DeviceStateAPI(mock_session).get_device_orchestrator_connections_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.connections_info_dataclass)

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_orchestrator_connections_info_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_orchestrator_connections_info(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_reboot_history(self, mock_session):
        # Arrange
        mock_session.endpoints = APIEndpointContainter(mock_session)
        mock_session.request.return_value = ManagerResponse(ResponseMock({"data": self.reboot_history}))
        # Act
        answer = DeviceStateAPI(mock_session).get_device_reboot_history(device_id="1.1.1.11")
        # Assert
        self.assertEqual(answer, self.reboot_history_payload)

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_reboot_history_empty(self, mock_session):
        # Arrange
        mock_session.endpoints = APIEndpointContainter(mock_session)
        mock_session.request.return_value = ManagerResponse(ResponseMock({"data": []}))
        # Act
        answer = DeviceStateAPI(mock_session).get_device_reboot_history(device_id="1.1.1.1")
        # Assert
        self.assertEqual(len(answer), 0)

    @patch("catalystwan.session.ManagerSession")
    def test_get_system_status(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.device
        # Act
        answer = DeviceStateAPI(mock_session).get_system_status(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.device_dataclass)

    @patch("catalystwan.session.ManagerSession")
    def test_get_system_status_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []

        # Act
        def answer():
            return DeviceStateAPI(mock_session).get_system_status(device_id="1.1.1.1")

        # Assert
        self.assertRaises(AssertionError, answer)

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_wan_interfaces(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.wan_interfaces
        # Act
        answer = DeviceStateAPI(mock_session).get_device_wan_interfaces(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.wan_interfaces_dataclass)

    @patch("catalystwan.session.ManagerSession")
    def test_get_device_wan_interfaces_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_device_wan_interfaces(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    def test_get_colors(self):
        pass  # TODO fix method before test

    def test_enable_data_stream(self):
        pass  # TODO fix method before test

    @patch("catalystwan.session.ManagerSession")
    def test_get_bfd_sessions(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.bfd_session
        # Act
        answer = DeviceStateAPI(mock_session).get_bfd_sessions(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, self.bfd_session_dataclass)

    @patch("catalystwan.session.ManagerSession")
    def test_get_bfd_sessions_empty(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = []
        # Act
        answer = DeviceStateAPI(mock_session).get_bfd_sessions(device_id="1.1.1.1")
        # Assert
        self.assertEqual(answer, [])

    @patch("catalystwan.session.ManagerSession")
    def test_wait_for_bfd_session_up(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.bfd_session
        # Act
        answer = DeviceStateAPI(mock_session).wait_for_bfd_session_up(system_ip="1.1.1.1")
        # Assert
        self.assertIsNone(answer)

    @patch("catalystwan.session.ManagerSession")
    def test_wait_for_bfd_session_up_timeout(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.bfd_session_down

        # Act
        def answer():
            return DeviceStateAPI(mock_session).wait_for_bfd_session_up(
                system_ip="1.1.1.1", sleep_seconds=1, timeout_seconds=1
            )

        # Assert
        self.assertRaises(RetryError, answer)

    @mark.skip(reason="10 minutes length")
    @patch("catalystwan.session.ManagerSession")
    def test_wait_for_device_state(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.device
        # Act
        answer = DeviceStateAPI(mock_session).wait_for_device_state(device_id="1.1.1.1")
        # Assert
        self.assertTrue(answer)

    @patch("catalystwan.session.ManagerSession")
    def test_wait_for_device_state_unreachable(self, mock_session):
        pass  # TODO fix method before test



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import ipaddress
from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.bool_str import BoolStr
from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator


class MetricType(str, Enum):
    TYPE1 = "type1"
    TYPE2 = "type2"


class Protocol(str, Enum):
    BGP = "bgp"
    CONNECTED = "connected"
    EIGRP = "eigrp"
    ISIS = "isis"
    LISP = "lisp"
    NAT_ROUTE = "nat-route"
    OMP = "omp"
    STATIC = "static"


class Redistribute(FeatureTemplateValidator):
    protocol: Protocol
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    dia: Optional[BoolStr] = True
    model_config = ConfigDict(populate_by_name=True)


class AdType(str, Enum):
    ON_STARTUP = "on-startup"


class RouterLsa(FeatureTemplateValidator):
    ad_type: AdType = Field(json_schema_extra={"vmanage_key": "ad-type"})
    time: int
    model_config = ConfigDict(populate_by_name=True)


class Translate(str, Enum):
    ALWAYS = "always"


class Network(str, Enum):
    BROADCAST = "broadcast"
    POINT_TO_POINT = "point-to-point"
    NON_BROADCAST = "non-broadcast"
    POINT_TO_MULTIPOINT = "point-to-multipoint"


class Type(str, Enum):
    MD5 = "md5"
    SHA1 = "sha1"


class Interface(FeatureTemplateValidator):
    name: str
    hello_interval: Optional[int] = Field(10, json_schema_extra={"vmanage_key": "hello-interval"})
    dead_interval: Optional[int] = Field(40, json_schema_extra={"vmanage_key": "dead-interval"})
    retransmit_interval: Optional[int] = Field(5, json_schema_extra={"vmanage_key": "retransmit-interval"})
    cost: Optional[int] = None
    network: Optional[Network] = Network.BROADCAST
    passive_interface: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "passive-interface"})
    type: Type = Field(json_schema_extra={"data_path": ["authentication"]})
    authentication_key: str = Field(
        json_schema_extra={"vmanage_key": "authentication-key", "data_path": ["authentication"]}
    )
    spi: Optional[int] = Field(default=None, json_schema_extra={"data_path": ["authentication", "ipsec"]})
    model_config = ConfigDict(populate_by_name=True)


class Range(FeatureTemplateValidator):
    address: ipaddress.IPv4Interface
    cost: Optional[int] = None
    no_advertise: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "no-advertise"})
    model_config = ConfigDict(populate_by_name=True)


class Area(FeatureTemplateValidator):
    a_num: int = Field(json_schema_extra={"vmanage_key": "a-num"})
    stub: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "no-summary", "data_path": ["stub"]}
    )
    nssa: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "no-summary", "data_path": ["nssa"]}
    )
    translate: Optional[Translate] = Field(default=None, json_schema_extra={"data_path": ["nssa"]})
    normal: Optional[BoolStr] = None
    interface: Optional[List[Interface]] = None
    range: Optional[List[Range]] = None
    model_config = ConfigDict(populate_by_name=True)


class RedistributeV6(FeatureTemplateValidator):
    protocol: Protocol
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    model_config = ConfigDict(populate_by_name=True)


class InterfaceV6(FeatureTemplateValidator):
    name: str
    hello_interval: Optional[int] = Field(10, json_schema_extra={"vmanage_key": "hello-interval"})
    dead_interval: Optional[int] = Field(40, json_schema_extra={"vmanage_key": "dead-interval"})
    retransmit_interval: Optional[int] = Field(5, json_schema_extra={"vmanage_key": "retransmit-interval"})
    cost: Optional[int] = None
    network: Optional[Network] = Network.BROADCAST
    passive_interface: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "passive-interface"})
    type: Type = Field(json_schema_extra={"data_path": ["authentication"]})
    authentication_key: str = Field(
        json_schema_extra={"vmanage_key": "authentication-key", "data_path": ["authentication"]}
    )
    spi: Optional[int] = Field(default=None, json_schema_extra={"data_path": ["authentication", "ipsec"]})
    model_config = ConfigDict(populate_by_name=True)


class RangeV6(FeatureTemplateValidator):
    address: ipaddress.IPv6Interface
    cost: Optional[int] = None
    no_advertise: Optional[bool] = Field(False, json_schema_extra={"vmanage_key": "no-advertise"})
    model_config = ConfigDict(populate_by_name=True)


class AreaV6(FeatureTemplateValidator):
    a_num: int = Field(json_schema_extra={"vmanage_key": "a-num"})
    stub: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "no-summary", "data_path": ["stub"]}
    )
    nssa: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "no-summary", "data_path": ["nssa"]}
    )
    translate: Optional[Translate] = Field(default=None, json_schema_extra={"data_path": ["nssa"]})
    normal: Optional[BoolStr] = None
    interface: Optional[List[InterfaceV6]] = None
    range: Optional[List[RangeV6]] = None
    model_config = ConfigDict(populate_by_name=True)


class CiscoOspfv3Model(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    router_id_v4: Optional[ipaddress.IPv4Address] = Field(
        default=None, json_schema_extra={"vmanage_key": "router-id", "data_path": ["ospfv3", "address-family", "ipv4"]}
    )
    reference_bandwidth_v4: Optional[int] = Field(
        100,
        json_schema_extra={
            "vmanage_key": "reference-bandwidth",
            "data_path": ["ospfv3", "address-family", "ipv4", "auto-cost"],
        },
    )
    rfc1583_v4: Optional[BoolStr] = Field(
        default=True,
        json_schema_extra={"vmanage_key": "rfc1583", "data_path": ["ospfv3", "address-family", "ipv4", "compatible"]},
    )
    originate_v4: Optional[BoolStr] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "originate",
            "data_path": ["ospfv3", "address-family", "ipv4", "default-information"],
        },
    )
    always_v4: Optional[BoolStr] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "always",
            "data_path": ["ospfv3", "address-family", "ipv4", "default-information", "originate"],
        },
    )
    metric_v4: Optional[int] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "metric",
            "data_path": ["ospfv3", "address-family", "ipv4", "default-information", "originate"],
        },
    )
    metric_type_v4: Optional[MetricType] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "metric-type",
            "data_path": ["ospfv3", "address-family", "ipv4", "default-information", "originate"],
        },
    )
    external_v4: Optional[int] = Field(
        110,
        json_schema_extra={
            "vmanage_key": "external",
            "data_path": ["ospfv3", "address-family", "ipv4", "distance-ipv4", "ospf"],
        },
    )
    inter_area_v4: Optional[int] = Field(
        110,
        json_schema_extra={
            "vmanage_key": "inter-area",
            "data_path": ["ospfv3", "address-family", "ipv4", "distance-ipv4", "ospf"],
        },
    )
    intra_area_v4: Optional[int] = Field(
        110,
        json_schema_extra={
            "vmanage_key": "intra-area",
            "data_path": ["ospfv3", "address-family", "ipv4", "distance-ipv4", "ospf"],
        },
    )
    delay_v4: Optional[int] = Field(
        200,
        json_schema_extra={
            "vmanage_key": "delay",
            "data_path": ["ospfv3", "address-family", "ipv4", "timers", "throttle", "spf"],
        },
    )
    initial_hold_v4: Optional[int] = Field(
        1000,
        json_schema_extra={
            "vmanage_key": "initial-hold",
            "data_path": ["ospfv3", "address-family", "ipv4", "timers", "throttle", "spf"],
        },
    )
    max_hold_v4: Optional[int] = Field(
        10000,
        json_schema_extra={
            "vmanage_key": "max-hold",
            "data_path": ["ospfv3", "address-family", "ipv4", "timers", "throttle", "spf"],
        },
    )
    distance_v4: Optional[int] = Field(
        110,
        json_schema_extra={
            "vmanage_key": "distance",
            "data_path": ["ospfv3", "address-family", "ipv4", "distance-ipv4"],
        },
    )
    name_v4: Optional[str] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "name", "data_path": ["ospfv3", "address-family", "ipv4", "table-map"]},
    )
    filter_v4: Optional[BoolStr] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "filter", "data_path": ["ospfv3", "address-family", "ipv4", "table-map"]},
    )
    redistribute_v4: Optional[List[Redistribute]] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "redistribute", "data_path": ["ospfv3", "address-family", "ipv4"]},
    )
    router_lsa_v4: Optional[List[RouterLsa]] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "router-lsa",
            "data_path": ["ospfv3", "address-family", "ipv4", "max-metric"],
        },
    )
    area_v4: Optional[List[Area]] = Field(
        default=None, json_schema_extra={"vmanage_key": "area", "data_path": ["ospfv3", "address-family", "ipv4"]}
    )
    router_id_v6: Optional[ipaddress.IPv4Address] = Field(
        default=None, json_schema_extra={"vmanage_key": "router-id", "data_path": ["ospfv3", "address-family", "ipv6"]}
    )
    reference_bandwidth_v6: Optional[int] = Field(
        100,
        json_schema_extra={
            "vmanage_key": "reference-bandwidth",
            "data_path": ["ospfv3", "address-family", "ipv6", "auto-cost"],
        },
    )
    rfc1583_v6: Optional[BoolStr] = Field(
        default=True,
        json_schema_extra={"vmanage_key": "rfc1583", "data_path": ["ospfv3", "address-family", "ipv6", "compatible"]},
    )
    originate_v6: Optional[BoolStr] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "originate",
            "data_path": ["ospfv3", "address-family", "ipv6", "default-information"],
        },
    )
    always_v6: Optional[BoolStr] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "always",
            "data_path": ["ospfv3", "address-family", "ipv6", "default-information", "originate"],
        },
    )
    metric_v6: Optional[int] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "metric",
            "data_path": ["ospfv3", "address-family", "ipv6", "default-information", "originate"],
        },
    )
    metric_type_v6: Optional[MetricType] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "metric-type",
            "data_path": ["ospfv3", "address-family", "ipv6", "default-information", "originate"],
        },
    )
    external_v6: Optional[int] = Field(
        110,
        json_schema_extra={
            "vmanage_key": "external",
            "data_path": ["ospfv3", "address-family", "ipv6", "distance-ipv6", "ospf"],
        },
    )
    inter_area_v6: Optional[int] = Field(
        110,
        json_schema_extra={
            "vmanage_key": "inter-area",
            "data_path": ["ospfv3", "address-family", "ipv6", "distance-ipv6", "ospf"],
        },
    )
    intra_area_v6: Optional[int] = Field(
        110,
        json_schema_extra={
            "vmanage_key": "intra-area",
            "data_path": ["ospfv3", "address-family", "ipv6", "distance-ipv6", "ospf"],
        },
    )
    delay_v6: Optional[int] = Field(
        200,
        json_schema_extra={
            "vmanage_key": "delay",
            "data_path": ["ospfv3", "address-family", "ipv6", "timers", "throttle", "spf"],
        },
    )
    initial_hold_v6: Optional[int] = Field(
        1000,
        json_schema_extra={
            "vmanage_key": "initial-hold",
            "data_path": ["ospfv3", "address-family", "ipv6", "timers", "throttle", "spf"],
        },
    )
    max_hold_v6: Optional[int] = Field(
        10000,
        json_schema_extra={
            "vmanage_key": "max-hold",
            "data_path": ["ospfv3", "address-family", "ipv6", "timers", "throttle", "spf"],
        },
    )
    distance_v6: Optional[int] = Field(
        110,
        json_schema_extra={
            "vmanage_key": "distance",
            "data_path": ["ospfv3", "address-family", "ipv6", "distance-ipv6"],
        },
    )
    name_v6: Optional[str] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "name", "data_path": ["ospfv3", "address-family", "ipv6", "table-map"]},
    )
    filter_v6: Optional[BoolStr] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "filter", "data_path": ["ospfv3", "address-family", "ipv6", "table-map"]},
    )
    redistribute_v6: Optional[List[RedistributeV6]] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "redistribute", "data_path": ["ospfv3", "address-family", "ipv6"]},
    )
    router_lsa_v6: Optional[List[RouterLsa]] = Field(
        default=None,
        json_schema_extra={
            "vmanage_key": "router-lsa",
            "data_path": ["ospfv3", "address-family", "ipv6", "max-metric"],
        },
    )
    area_v6: Optional[List[AreaV6]] = Field(
        default=None, json_schema_extra={"vmanage_key": "area", "data_path": ["ospfv3", "address-family", "ipv6"]}
    )

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cisco_ospfv3"



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.models.policy.policy_definition import PolicyDefinitionBase


class Region(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    name: str
    site_lists: List[UUID] = Field(validation_alias="siteLists", serialization_alias="siteLists")


class MeshPolicyDefinition(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    vpn_list: UUID = Field(validation_alias="vpnList", serialization_alias="vpnList")
    regions: List[Region] = []


class MeshPolicy(PolicyDefinitionBase):
    model_config = ConfigDict(populate_by_name=True)
    type: Literal["mesh"] = "mesh"
    definition: MeshPolicyDefinition

    @staticmethod
    def from_vpn_list(name: str, vpn_list: UUID) -> "MeshPolicy":
        return MeshPolicy(name=name, definition=MeshPolicyDefinition(vpn_list=vpn_list))

    def add_region(self, name: str, site_lists: List[UUID]) -> Region:
        region = Region(name=name, site_lists=site_lists)
        self.definition.regions.append(region)
        return region



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import tempfile
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch
from uuid import uuid4

from packaging.version import Version  # type: ignore

from catalystwan.api.task_status_api import Task
from catalystwan.api.tenant_migration_api import ImportTask, TenantMigrationAPI
from catalystwan.endpoints.tenant_migration import ImportInfo, MigrationInfo
from catalystwan.models.tenant import Tenant, TenantExport


class TestTenantMigrationAPI(unittest.TestCase):
    @patch("catalystwan.session.ManagerSession")
    def setUp(self, session_mock):
        self.session = session_mock
        self.api = TenantMigrationAPI(self.session)

    def test_export_tenant(self):
        tenant = Tenant(desc="Test Tenant", name="test_tenant", subdomain="test_subdomain", org_name="test_org")
        task = self.api.export_tenant(tenant=tenant)
        self.assertIsInstance(task, Task)

    def test_export_migration_tenant(self):
        tenant = TenantExport(
            name="test_tenant",
            desc="Test Tenant Description",
            subdomain="test_subdomain",
            org_name="test_org",
            is_destination_overlay_mt=False,
            migration_key="Cisco12345",
        )
        task = self.api.export_tenant(tenant=tenant)
        self.assertIsInstance(task, Task)

    def test_download(self):
        content = b"\xFFtest_data"
        with tempfile.TemporaryDirectory() as tmpdir:
            download_path = Path(tmpdir) / "test.tar.gz"
            self.session.endpoints.tenant_migration.download_tenant_data = MagicMock(return_value=content)
            self.api.download(download_path)
            assert open(download_path, "rb").read() == content

    def test_import_tenant(self):
        self.session.api_version = Version("20.12")
        migration_key = "mkey1"
        migration_id = str(uuid4())
        with tempfile.TemporaryDirectory() as tmpdir:
            import_file = Path(tmpdir) / "tenant.tar.gz"
            with open(import_file, "wb") as f:
                f.write(b"\xFEtest_data")
            self.session.endpoints.tenant_migration.import_tenant_data = MagicMock(
                return_value=ImportInfo(
                    process_id="123",
                    migration_token_url=("/dataservice/tenantmigration/migrationToken?" f"migrationId={migration_id}"),
                )
            )
            task = self.api.import_tenant(import_file, migration_key)
            self.assertIsInstance(task, ImportTask)
            assert task.import_info.migration_token_query_params.migration_id == migration_id

    def test_import_tenant_with_key(self):
        self.session.api_version = Version("20.13")
        migration_key = "mkey2"
        migration_id = str(uuid4())
        with tempfile.TemporaryDirectory() as tmpdir:
            import_file = Path(tmpdir) / "tenant.tar.gz"
            with open(import_file, "wb") as f:
                f.write(b"\xFEtest_data")
            self.session.endpoints.tenant_migration.import_tenant_data_with_key = MagicMock(
                return_value=ImportInfo(
                    process_id="123",
                    migration_token_url=("/dataservice/tenantmigration/migrationToken?" f"migrationId={migration_id}"),
                )
            )
            task = self.api.import_tenant(import_file, migration_key)
            self.assertIsInstance(task, ImportTask)
            assert task.import_info.migration_token_query_params.migration_id == migration_id

    def test_store_token(self):
        token_data = "test_token1987"
        with tempfile.TemporaryDirectory() as tmpdir:
            download_path = Path(tmpdir) / "token.txt"
            migration_id = "123"
            self.session.endpoints.tenant_migration.get_migration_token = MagicMock(return_value=token_data)
            self.api.store_token(migration_id, download_path)
            assert open(download_path, "r").read() == token_data

    def test_migrate_network(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            token_file = Path(tmpdir) / "token_file.txt"
            with open(token_file, "wb") as f:
                f.write(b"token_text")
            self.api.session.endpoints.tenant_migration.migrate_network = MagicMock(
                return_value=MigrationInfo(processId="123")
            )
            task = self.api.migrate_network(token_file)
            self.assertIsInstance(task, Task)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import datetime
from typing import List, Literal, Optional, Sequence, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field


class PolicyId(BaseModel):
    policy_id: UUID = Field(serialization_alias="policyId", validation_alias="policyId")


class NGFirewallZoneListEntry(BaseModel):
    src_zone_list_id: UUID = Field(serialization_alias="srcZoneListId", validation_alias="srcZoneListId")
    dst_zone_list_id: UUID = Field(serialization_alias="dstZoneListId", validation_alias="dstZoneListId")
    model_config = ConfigDict(populate_by_name=True)


class AssemblyItemBase(BaseModel):
    definition_id: UUID = Field(serialization_alias="definitionId", validation_alias="definitionId")
    type: str
    model_config = ConfigDict(populate_by_name=True)


class ZoneBasedFWAssemblyItem(AssemblyItemBase):
    type: Literal["zoneBasedFW"] = "zoneBasedFW"


class NGFirewallAssemblyItem(AssemblyItemBase):
    type: Literal["zoneBasedFW"] = "zoneBasedFW"
    entries: List[NGFirewallZoneListEntry] = []

    def add_zone_pair(self, src_zone_id: UUID, dst_zone_id: UUID):
        self.entries.append(NGFirewallZoneListEntry(src_zone_list_id=src_zone_id, dst_zone_list_id=dst_zone_id))


class DNSSecurityAssemblyItem(AssemblyItemBase):
    type: Literal["DNSSecurity"] = "DNSSecurity"


class IntrusionPreventionAssemblyItem(AssemblyItemBase):
    type: Literal["intrusionPrevention"] = "intrusionPrevention"


class URLFilteringAssemblyItem(AssemblyItemBase):
    type: Literal["urlFiltering"] = "urlFiltering"


class AdvancedMalwareProtectionAssemblyItem(AssemblyItemBase):
    type: Literal["advancedMalwareProtection"] = "advancedMalwareProtection"


class SSLDecryptionAssemblyItem(AssemblyItemBase):
    type: Literal["sslDecryption"] = "sslDecryption"


class PolicyDefinition(BaseModel):
    assembly: Sequence[AssemblyItemBase] = []


class PolicyCreationPayload(BaseModel):
    policy_name: str = Field(
        serialization_alias="policyName",
        validation_alias="policyName",
        pattern="^[a-zA-Z0-9_-]{1,127}$",
        description="Can include only alpha-numeric characters, hyphen '-' or underscore '_'; maximum 127 characters",
    )
    policy_description: str = Field(
        default="default description", serialization_alias="policyDescription", validation_alias="policyDescription"
    )
    policy_type: str = Field(serialization_alias="policyType", validation_alias="policyType")
    policy_definition: Union[PolicyDefinition, str] = Field(
        serialization_alias="policyDefinition",
        validation_alias="policyDefinition",
    )
    is_policy_activated: bool = Field(
        default=False, serialization_alias="isPolicyActivated", validation_alias="isPolicyActivated"
    )
    model_config = ConfigDict(populate_by_name=True)


class PolicyEditPayload(PolicyCreationPayload, PolicyId):
    pass


class PolicyInfo(PolicyEditPayload):
    created_by: str = Field(serialization_alias="createdBy", validation_alias="createdBy")
    created_on: datetime.datetime = Field(serialization_alias="createdOn", validation_alias="createdOn")
    last_updated_by: str = Field(serialization_alias="lastUpdatedBy", validation_alias="lastUpdatedBy")
    last_updated_on: datetime.datetime = Field(serialization_alias="lastUpdatedOn", validation_alias="lastUpdatedOn")
    policy_version: Optional[str] = Field(None, serialization_alias="policyVersion", validation_alias="policyVersion")


class PolicyPreview(BaseModel):
    preview: str



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from datetime import datetime as dt
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from urllib.parse import quote

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class LogsAPI:
    """Class for get and save audit log file

    Usage example:
        # Create session
        session = create_manager_session(...)
        # Get auditlogs
        session.api.logs.get_auditlogs()
    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session

    def get_auditlogs(self, file_path: Optional[str] = None, n_hours: int = 1) -> None:
        query = {
            "query": {
                "condition": "AND",
                "rules": [
                    {
                        "field": "entry_time",
                        "type": "date",
                        "value": [f"{n_hours}"],
                        "operator": "last_n_hours",
                    }
                ],
            }
        }
        addon_to_url = quote(str(query), safe="").replace("%27", "%22")

        url_path = f"/dataservice/auditlog/severity?query={addon_to_url}"
        logs = self.session.get_data(url_path)

        if file_path is None:
            file_path = str(Path(__file__).parents[0] / "audit.log")

        with open(file_path, "w") as file:
            for log in logs:
                time = dt.utcfromtimestamp(log["entry_time"] / 1000)
                time_readable = time.strftime("%Y-%m-%d %H:%M:%S")
                file.write(
                    f"Entry time: {time_readable} - LogId: {log['logid']} - "
                    f"Log message: {log['logmessage']} - TenantId: {log['tenant']}\n"
                )

        logger.info(f"Logs saved to {file_path}")

        return None



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Union

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class ApplicationListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    app_list: Global[str] = Field(serialization_alias="app", validation_alias="app")


class ApplicationFamilyListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    app_list_family: Global[str] = Field(serialization_alias="appFamily", validation_alias="appFamily")


class ApplicationListParcel(_ParcelBase):
    entries: List[Union[ApplicationListEntry, ApplicationFamilyListEntry]] = Field(
        default=[], validation_alias=AliasPath("data", "entries")
    )

    def add_application(self, application: str):
        self.entries.append(ApplicationListEntry(app_list=as_global(application)))

    def add_application_family(self, application_family: str):
        self.entries.append(ApplicationFamilyListEntry(app_list_family=as_global(application_family)))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum
from typing import Any, Dict, Generic, Literal, Optional, TypeVar, get_origin

from pydantic import AliasPath, BaseModel, ConfigDict, Field, PrivateAttr, model_serializer

T = TypeVar("T")


class _ParcelBase(BaseModel):
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True, populate_by_name=True)
    parcel_name: str = Field(
        min_length=1,
        max_length=128,
        pattern=r'^[^&<>! "]+$',
        serialization_alias="name",
        validation_alias="name",
    )
    parcel_description: Optional[str] = Field(
        default=None,
        serialization_alias="description",
        validation_alias="description",
        description="Set the parcel description",
    )
    data: Optional[Any] = None
    _parcel_data_key: str = PrivateAttr(default="data")

    @model_serializer(mode="wrap")
    def envelope_parcel_data(self, handler) -> Dict[str, Any]:
        model_dict = handler(self)
        model_dict[self._parcel_data_key] = {}
        remove_keys = []

        for key in model_dict.keys():
            field_info = self.model_fields.get(key)
            if field_info and isinstance(field_info.validation_alias, AliasPath):
                aliases = field_info.validation_alias.convert_to_aliases()
                if aliases and aliases[0] == self._parcel_data_key and len(aliases) == 2:
                    model_dict[self._parcel_data_key][aliases[1]] = model_dict[key]
                    remove_keys.append(key)
        for key in remove_keys:
            del model_dict[key]
        return model_dict


class OptionType(str, Enum):
    GLOBAL = "global"
    DEFAULT = "default"
    VARIABLE = "variable"


class ParcelAttribute(BaseModel):
    model_config = ConfigDict(extra="forbid")
    option_type: OptionType = Field(serialization_alias="optionType", validation_alias="optionType")


# https://github.com/pydantic/pydantic/discussions/6090
# Usage: Global[str](value="test")
class Global(ParcelAttribute, Generic[T]):
    option_type: OptionType = Field(
        default=OptionType.GLOBAL, serialization_alias="optionType", validation_alias="optionType"
    )
    value: T

    def __len__(self) -> int:
        if isinstance(self.value, (str, list)):
            return len(self.value)
        return -1

    def __ge__(self, other: Any) -> bool:
        if isinstance(self.value, int):
            return self.value >= other
        return False

    def __le__(self, other: Any) -> bool:
        if isinstance(self.value, int):
            return self.value <= other
        return False


class Variable(ParcelAttribute):
    option_type: OptionType = Field(
        default=OptionType.VARIABLE, serialization_alias="optionType", validation_alias="optionType"
    )
    value: str = Field(pattern=r"^\{\{[.\/\[\]a-zA-Z0-9_-]+\}\}$", min_length=1, max_length=64)


class Default(ParcelAttribute, Generic[T]):
    option_type: OptionType = Field(
        default=OptionType.DEFAULT, serialization_alias="optionType", validation_alias="optionType"
    )
    value: Any


def as_global(value: Any, generic_alias: Any = None):
    """Produces Global object given only value (type is induced from value)

    Args:
        value (Any): value of Global object to be produced
        generic_alias (Any, optional): specify alias type like Literal. Defaults to None.

    Returns:
        Global[Any]: global option type object
    """
    if generic_alias is None:
        return Global[type(value)](value=value)  # type: ignore
    elif get_origin(generic_alias) is Literal:
        return Global[generic_alias](value=value)  # type: ignore
    TypeError("Inappropriate type for argument generic_alias")


def as_variable(value: str):
    """Produces Variable object from variable name string

    Args:
        value (str): value of Variable object to be produced

    Returns:
        Variable: variable option type object
    """
    return Variable(value=value)


def as_default(value: Any, generic_alias: Any = None):
    """Produces Default object given only value (type is induced from value)

    Args:
        value (Any): value of Default object to be produced
        generic_alias (Any, optional): specify alias type like Literal. Defaults to None.

    Returns:
        Default[Any]: default option type object
    """
    if generic_alias is None:
        return Default[type(value)](value=value)  # type: ignore
    elif get_origin(generic_alias) is Literal:
        return Default[generic_alias](value=value)  # type: ignore
    TypeError("Inappropriate type for argument generic_alias")



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set

from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed  # type: ignore

from catalystwan.dataclasses import AlarmData
from catalystwan.typed_list import DataSequence
from catalystwan.utils.creation_tools import create_dataclass, flatten_dict

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

logger = logging.getLogger(__name__)


class AlarmsAPI:
    """API methods of vManage for Alarms.

    Attributes:
        session: logged in API client session
    """

    URL = "/dataservice/alarms"

    def __init__(self, session: ManagerSession):
        self.session = session

    def get(self, from_time: Optional[int] = None) -> DataSequence[AlarmData]:
        """Data sequence of alarms.

        Args:
            from_time: Gets alarms from time in hour. Defaults to None - gets all alrams.

        Returns:
            DataSequence[AlarmData] of getted alarms.

        Examples:
            Get all alarms:
            >>> alarms = AlarmsAPI(session).get()

            Get all alarms from 3 hours:
            >>> alarms = AlarmsAPI(session).get(from_time=3)

            Get all not viewed alarms:
            >>> alarms = AlarmsAPI(session).get()
            >>> not_viewed_alarms = alarms.filter(viewed=False)

            Get all critical alarms:
            >>> alarms = AlarmsAPI(session).get()
            >>> critical_alarms = alarms.filter(severity=Severity.CRITICAL)
        """
        query: Dict[str, Any] = {
            "query": {
                "condition": "AND",
                "rules": [{"field": "active", "type": "boolean", "value": ["true"], "operator": "equal"}],
            }
        }
        if from_time:
            query["query"]["rules"].append(
                {
                    "value": [str(from_time)],
                    "field": "entry_time",
                    "type": "date",
                    "operator": "last_n_hours",
                }
            )

        response = self.session.post(url=AlarmsAPI.URL, json=query).json()["data"]
        alarms = [create_dataclass(AlarmData, flatten_dict(alarm)) for alarm in response]
        logger.info("Current alarms collected successfully.")

        return DataSequence(AlarmData, alarms)

    def mark_all_as_viewed(self) -> None:
        """Marks all alarms as viewed."""

        self.session.post(f"{AlarmsAPI.URL}/markallasviewed")
        logger.info("Alarms mark as viewed.")

    def check_alarms(
        self,
        expected: List[Dict[str, str]],
        timeout_seconds: int = 240,
        sleep_seconds: int = 5,
    ) -> Dict[str, set]:
        """Checks if alarms have occurred.

        For a specified period of time, the method queries the API about unread alarms and
        checks if they occur in the expected ones. If all alarms occur before the time expires,
        it will not poll any further.

        Args:
          expected(list): The list of expected alarms.
          timeout_seconds (int): Failure timeout.
          sleep_seconds (int): Sleep time.

        Returns:
          The dictionary with alarms that occurred (key 'found') and did not (key 'no-found')

        Examples:
            >>> expected_alarms = [{
                "type": "memory-usage",
                "rulename": "memory-usage",
                "component": "System",
                "severity": "Medium",
                "system_ip": "192.168.1.2",
                "acknowledged": true,
                "active": true,
                }]
            >>> result = AlarmsAPI(session).check_alarms(expected = expected_alarms)
        """

        alarms_expected = {create_dataclass(AlarmData, flatten_dict(expected_alarm)) for expected_alarm in expected}

        verification = AlarmVerification(logger, self.get)
        verification.verify(alarms_expected, timeout_seconds, sleep_seconds)

        logger.info(f"found alarms: {verification.found}")
        logger.info(f"not-found alarms: {verification.not_found}")

        return {"found": verification.found, "not-found": verification.not_found}


class AlarmVerification:
    """The Alarms verification class

    Attributes:
        alarms_getter (Callable): The function that returns a set of current alarms.
        logger (Logger): Logger.
    """

    def __init__(self, logger: logging.Logger, alarms_getter: Callable[[], DataSequence[AlarmData]]) -> None:
        self.alarms_getter = alarms_getter
        self.logger = logger

    def __check(self, expected: AlarmData) -> bool:
        """The checking if the expected alarm is included in the alarm set.

        Args:
          expected(AlarmData): The expected alarm.

        Returns:
          True if the requested alarm appears in the current alarm set.
        """
        not_viewed_alarms = self.alarms_getter().filter(viewed=False)
        checked = {expected.lowercase().issubset(actual.lowercase()) for actual in not_viewed_alarms}
        return any(checked)

    def verify(self, expected: Set[AlarmData], timeout_seconds: int, sleep_seconds: int):
        """The verifying if the expected alarms is included in the actual alarms set.

        Args:
          expected(Set[AlarmData]): The set expected alarms.
          timeout_seconds (int): Failure timeout.
          sleep_seconds (int): Sleep time.

        """
        self.found: Set[AlarmData] = set()
        self.not_found: Set[AlarmData] = set()

        def _log_exception(retry_state):
            self.logger.error(f"Cannot found alarms in {timeout_seconds}.")
            self.logger.error(f"Original exception: {retry_state.outcome.exception()}.")

        def check(founds):
            return expected != founds

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check),
            retry_error_callback=_log_exception,
        )
        def wait_for_alarms():
            self.logger.info(f"waiting for alarms: {expected ^ self.found}")
            for expected_alarm in expected ^ self.found:
                if self.__check(expected_alarm):
                    self.found.add(expected_alarm)
            self.not_found = expected ^ self.found
            return self.found

        wait_for_alarms()



# Copyright 2022 Cisco Systems, Inc. and its affiliates

"""
Module for handling admintech logs for a device
"""
from __future__ import annotations

import logging
import time
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from catalystwan.dataclasses import AdminTech, DeviceAdminTech
from catalystwan.exceptions import CatalystwanException, ManagerHTTPError
from catalystwan.response import ManagerResponse
from catalystwan.utils.creation_tools import create_dataclass

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

logger = logging.getLogger(__name__)


class GenerateAdminTechLogError(CatalystwanException):
    pass


class DownloadAdminTechLogError(CatalystwanException):
    pass


class RequestTokenIdNotFound(CatalystwanException):
    pass


class AdminTechAPI:
    """Class for handling admintech logs for a device.

    Attributes:
        session: logged in API client session

    Usage example:
        # Create session
        session = create_manager_session(...)
        # Get admintech data for all devices
        all_admintechs = session.api.admin_tech.get_all()
    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session

    def __str__(self) -> str:
        return str(self.session)

    def get(self, device_id: str) -> List[DeviceAdminTech]:
        """Gets admintech log information for a device.

        Args:
            device_id: device ID (usually system-ip)
        Returns:
            AdminTech object list for given device
        """
        body = {"deviceIP": device_id}
        response = self.session.post(url="/dataservice/device/tools/admintechlist", json=body)
        items = response.json()["data"]
        return [create_dataclass(DeviceAdminTech, item) for item in items]

    def get_all(self) -> List[AdminTech]:
        """Gets admintech log information for all devices.

        Returns:
            AdminTech objects list for all devices
        """
        response = self.session.get("/dataservice/device/tools/admintechs")
        items = response.json()["data"]
        return [create_dataclass(AdminTech, item) for item in items]

    def generate(
        self,
        device_id: str,
        exclude_cores: bool = True,
        exclude_tech: bool = False,
        exclude_logs: bool = True,
        request_timeout: int = 3600,
        polling_timeout: int = 1200,
        polling_interval: int = 30,
    ) -> str:
        """Generates admintech log for a device.
        Args:
            device_id: device ID (usually system-ip)
            exclude_cores: exclude core in generated admintech log file
            exclude_tech: exclude tech in generated admintech log file
            exclude_logs: exclude logs in generated admintech log file
            request_timeout: wait time in seconds to generate admintech after request
            polling_timeout: retry period in seconds for successfull request
            polling_interval: polling interval in seconds between request attempts
        Returns:
            filename of generated admintech log
        """
        create_admin_tech_error_msgs = "Admin tech creation already in progress"
        body = {
            "deviceIP": device_id,
            "exclude-cores": exclude_cores,
            "exclude-tech": exclude_tech,
            "exclude-logs": exclude_logs,
        }
        polling_timer = polling_timeout
        while polling_timer > 0:
            logger.info(
                f"Starting AdminTech log creation for {device_id}, waiting up to {request_timeout} seconds to complete"
            )
            try:
                response = self.session.post(
                    url="/dataservice/device/tools/admintech",
                    json=body,
                    timeout=request_timeout,
                )
            except ManagerHTTPError as http_error:
                response = http_error.response  # type: ignore
            if response.status_code == 200:
                return response.json()["fileName"]
            if response.status_code == 400 and create_admin_tech_error_msgs in response.json().get("error", {}).get(
                "details", ""
            ):
                logger.warning(f"Admin tech creation already in progress, retrying in {polling_interval} seconds")
            else:
                raise GenerateAdminTechLogError(f"It is not possible to generate admintech log for {device_id}")
            time.sleep(polling_interval)
            polling_timer -= polling_interval
        raise GenerateAdminTechLogError(f"It is not possible to generate admintech log for {device_id}")

    def _get_token_id(self, filename: str, timeout: int, interval: int) -> str:
        # Wait for the file to be ready and obtain the token_id
        end_time = time.time() + timeout
        while time.time() < end_time:
            admin_techs = self.get_all()
            for admin_tech in admin_techs:
                if filename == admin_tech.filename and admin_tech.state == "done":
                    return admin_tech.token_id
            time.sleep(interval)
        raise RequestTokenIdNotFound(
            f"requestTokenId of admin tech generation request not found for file name: {filename}"
        )

    def delete(self, filename: str, timeout: int = 3600, interval: int = 30) -> ManagerResponse:
        """Deletes admin tech logs for a device.
        Args:
            filename: name of admin_tech file
        Returns:
            response: http response for delete operation
        """

        token_id = self._get_token_id(filename, timeout, interval)
        response = self.session.delete(f"/dataservice/device/tools/admintech/{token_id}")
        if response.status_code == 200:
            logger.info(f"Deleted AdminTech file {filename} on remote")
        return response

    def download(self, filename: str, download_dir: Optional[Path] = None) -> Path:
        """Downloads admintech log for a device.
        Args:
            filename: name of admin_tech file
            download_dir: download directory (defaults to current working directory)
        Returns:
            path to downloaded admin_tech file
        """
        if not download_dir:
            download_dir = Path.cwd()
        download_path = download_dir / filename
        url = f"/dataservice/device/tools/admintech/download/{filename}"
        if self.session.get_file(url=url, filename=download_path).status_code != 200:
            raise DownloadAdminTechLogError(f"Cannot download admin tech file: {filename} from remote")
        logger.info(f"Downloaded AdminTech file to: {download_path}")
        return download_path



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# from typing import Dict, Generic, Type, TypeVar

# from vmngclient.api.configuration_groups.parcel import Parcel

# ParcelModel = TypeVar("ParcelModel", bound=Parcel)


# class Builder(Generic[ParcelModel]):
#     model: Type[ParcelModel]
#     values: Dict[str, object]

#     def __init__(self, model: Type[ParcelModel]) -> None:
#         super().__setattr__("model", model)
#         super().__setattr__("values", {})

#     def __setattr__(self, name: str, value: object) -> None:
#         self.values[name] = value

#     def build(self) -> ParcelModel:
#         return self.model(**self.values)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"

from enum import Enum
from typing import List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, delete, get, post, put
from catalystwan.typed_list import DataSequence


class ImageType(str, Enum):
    SOFTWARE = "software"
    CONTAINER = "container"
    LXC = "lxc"
    VIRTUALMACHINE = "virtualmachine"
    VIRTUALMACHINE_DISKIMG = "virtualmachine-diskimg"
    VIRTUALMACHINE_SCAFFOLD = "virtualmachine-scaffold"
    UTDSIGNATURE = "utdsignature"
    UTDSIGNATUREIPS = "utdsignatureips"
    UTDSIGNATURECUSTOM = "utdsignaturecustom"
    WAAS = "waas"
    DOCKERTYPE = "dockertype"
    FIRMWARE = "firmware"
    RUMREPORT = "rumreport"
    LICENSEFILE = "licensefile"
    ACKFILE = "ackfile"
    NA = "na"


class SoftwareImageQuery(BaseModel):
    image_type: ImageType = Field(default=ImageType.SOFTWARE, alias="imageType")
    vnf_type: Optional[ImageType] = Field(default=None, alias="vnfType")


class RemoteServerProtocol(str, Enum):
    FTP = "FTP"
    HTTP = "HTTP"
    SCP = "SCP"


class RemoteServer(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    image_location_prefix: str = Field(
        default="/", serialization_alias="imageLocationPrefix", validation_alias="imageLocationPrefix"
    )
    remote_server_name: str = Field(serialization_alias="remoteServerName", validation_alias="remoteServerName")
    remote_server_password: str = Field(
        serialization_alias="remoteServerPassword", validation_alias="remoteServerPassword"
    )
    remote_server_port: int = Field(
        default=21, serialization_alias="remoteServerPort", validation_alias="remoteServerPort"
    )
    remote_server_protocol: RemoteServerProtocol = Field(
        default=RemoteServerProtocol.FTP,
        serialization_alias="remoteServerProtocol",
        validation_alias="remoteServerProtocol",
    )
    remote_server_url: str = Field(serialization_alias="remoteServerUrl", validation_alias="remoteServerUrl")
    remote_server_user: str = Field(serialization_alias="remoteServerUser", validation_alias="remoteServerUser")
    remote_server_vpn: int = Field(
        ge=0, le=65527, serialization_alias="remoteServerVPN", validation_alias="remoteServerVPN"
    )


class RemoteServerInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    image_location_prefix: Optional[str] = Field(
        default=None, serialization_alias="imageLocationPrefix", validation_alias="imageLocationPrefix"
    )
    remote_server_id: Optional[str] = Field(
        default=None, serialization_alias="remoteServerId", validation_alias="remoteServerId"
    )
    remote_server_name: Optional[str] = Field(
        default=None, serialization_alias="remoteServerName", validation_alias="remoteServerName"
    )
    remote_server_port: Optional[int] = Field(
        default=None, serialization_alias="remoteServerPort", validation_alias="remoteServerPort"
    )
    remote_server_protocol: Optional[RemoteServerProtocol] = Field(
        default=None, serialization_alias="remoteServerProtocol", validation_alias="remoteServerProtocol"
    )
    remote_server_url: Optional[str] = Field(
        default=None, serialization_alias="remoteServerUrl", validation_alias="remoteServerUrl"
    )
    remote_server_user: Optional[str] = Field(
        default=None, serialization_alias="remoteServerUser", validation_alias="remoteServerUser"
    )
    remote_server_vpn: Optional[int] = Field(
        default=None, serialization_alias="remoteServerVPN", validation_alias="remoteServerVPN"
    )
    updated_on: Optional[int] = Field(default=None, serialization_alias="updatedOn", validation_alias="updatedOn")


class SoftwareRemoteServer(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    filename: str = Field(default=None, serialization_alias="fileName", validation_alias="fileName")
    remote_server_id: str = Field(default=None, serialization_alias="remoteServerId", validation_alias="remoteServerId")
    smu_defect_id: Optional[str] = Field(
        default=None, serialization_alias="smuDefectId", validation_alias="smuDefectId"
    )
    smu_type: Optional[str] = Field(default=None, serialization_alias="smuType", validation_alias="smuType")


class SoftwareImageDetails(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    arch: Optional[str] = Field(default=None)
    available_files: Optional[str] = Field(
        default=None, serialization_alias="availableFiles", validation_alias="availableFiles"
    )
    available_smu_versions: Optional[Union[List[str], str]] = Field(
        default=None, serialization_alias="availableSmuVersions", validation_alias="availableSmuVersions"
    )
    checksum_map: Optional[str] = Field(default=None, serialization_alias="checksumMap", validation_alias="checksumMap")
    controller_version_name: Optional[str] = Field(
        default=None, serialization_alias="controllerVersionName", validation_alias="controllerVersionName"
    )
    cw_version: Optional[str] = Field(default=None, serialization_alias="cwVersion", validation_alias="cwVersion")
    file_name_map: Optional[str] = Field(
        default=None, serialization_alias="fileNameMap", validation_alias="fileNameMap"
    )
    image_properties_json: Optional[str] = Field(
        default=None, serialization_alias="imagePropertiesJson", validation_alias="imagePropertiesJson"
    )
    image_type: Optional[str] = Field(default=None, serialization_alias="imageType", validation_alias="imageType")
    network_function_type: Optional[str] = Field(
        default=None, serialization_alias="networkFunctionType", validation_alias="networkFunctionType"
    )
    platform_family: Optional[List[str]] = Field(
        default=None, serialization_alias="platformFamily", validation_alias="platformFamily"
    )
    rid: Optional[int] = Field(default=None, serialization_alias="@rid", validation_alias="@rid")
    smu_compatible_version: Optional[str] = Field(
        default=None, serialization_alias="smuCompatibleVersion", validation_alias="smuCompatibleVersion"
    )
    system_properties_xml: Optional[str] = Field(
        default=None, serialization_alias="systemPropertiesXml", validation_alias="systemPropertiesXml"
    )
    updated_on: Optional[int] = Field(default=None, serialization_alias="updatedOn", validation_alias="updatedOn")
    vendor: Optional[str] = Field(default=None)
    version_id: Optional[str] = Field(default=None, serialization_alias="versionId", validation_alias="versionId")
    version_id_map: Optional[str] = Field(
        default=None, serialization_alias="versionIdMap", validation_alias="versionIdMap"
    )
    version_name: Optional[str] = Field(default=None, serialization_alias="versionName", validation_alias="versionName")
    version_type: Optional[str] = Field(default=None, serialization_alias="versionType", validation_alias="versionType")
    version_type_name: Optional[str] = Field(
        default=None, serialization_alias="versionTypeName", validation_alias="versionTypeName"
    )
    version_url: Optional[str] = Field(default=None, serialization_alias="versionURL", validation_alias="versionURL")
    vnf_properties_json: Optional[str] = Field(
        default=None, serialization_alias="vnfPropertiesJson", validation_alias="vnfPropertiesJson"
    )
    remote_server_id: str = Field(default=None, serialization_alias="remoteServerId", validation_alias="remoteServerId")


class ConfigurationSoftwareActions(APIEndpoints):
    @get("/device/action/remote-server", "data")
    def get_list_of_remote_servers(self) -> DataSequence[RemoteServerInfo]:
        ...

    @post("/device/action/remote-server")
    def add_new_remote_server(self, payload: RemoteServer) -> None:
        ...

    @get("/device/action/remote-server/{id}")
    def get_remote_server(self, id: UUID) -> RemoteServerInfo:
        ...

    @put("/device/action/remote-server/{id}")
    def update_remote_server(self, id: UUID, payload: SoftwareRemoteServer) -> None:
        ...

    @delete("/device/action/remote-server/{id}")
    def remove_remote_server(self, id: UUID) -> None:
        ...

    @get("/device/action/software", "data")
    def get_software_images(self) -> DataSequence[SoftwareImageDetails]:
        ...

    @post("/device/action/software")
    def upload_software_from_remote_server(self, payload: SoftwareRemoteServer) -> None:
        ...

    @delete("/device/action/software/{version_id}")
    def delete_software_from_software_repository(self, version_id: UUID) -> None:
        ...

    @get("/device/action/software/images", "data")
    def get_list_of_all_images(self, params: SoftwareImageQuery) -> DataSequence[SoftwareImageDetails]:
        ...
